"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

import os
import subprocess as sp
from numpy import *
from datetime import datetime, timedelta
from builtins import input

from HWS.HS_CamConfig import *
from HWS.HS_Image import *


class HS_Camera:
    """A class to operate a DALSA Pantera 1M60 Camera through a frame grabber.

    This class should be a single communication point between the Hartmann
    Sensor equipments (camera and frame grabber) and the rest of the codes.

    It holds variables and methods relevant to the system initialization
    and diagnostics, as well as image acquisition from the frame grabber.

    **Instantiation**

    During instantiation, if the input parmeter ``check_status`` is ``True``,
    the class checks if the connection can be made to the camera assumed
    connected to the device channel specified by the instance variable
    ``device_no``. If the connection cannot be made, it raises an error and
    exits.

    After a successful connection, the class creates an instance of the
    class :class:`HS_CamConfig <HWS.HS_CamConfig.HS_CamConfig>` and
    assigns it to the instance variable ``camera_parameters``.

    Then if the instance variable ``camera_folder`` is not set by a user,
    :func:`~HS_Camera.set_camera_folder` is called to set a default folder for
    storing acquired image frames and other data.

    **Camera Commands**

    All of the Dalsa Pantera 1M60 camera commands are represented in
    this class, although not all of them may be useful or relevant for a
    normal day-to-day operation of the camera.

    As such the descriptions of those methods are not full, and a user should
    consult the camera manual for the details.

    To call the camera commands from Python, some methods use the method
    :func:`~HS_Camera.serial_cmd` and other methods (which requires additional
    parameters) use :func:`~HS_Camera.serial_cmd_alt`. This is because the
    methods with parameters do not run when passed as a *list* (as was
    recommended by Python) to the medhos in the ``subprocess`` module.

    To overcome this, the camera commands with parameters are called by
    passing the command and its parameters as a whole string, which is what
    :func:`~HS_Camera.serial_cmd_alt` does.

    **Instance Variables**

    - ``device_no``: the number 0 or 1 specifying the frame grabber's device
      channel to which a camera is connected.

      :Type: *integer* of value 0 or 1

    - ``fg_directory``: the directory in which the frame grabber is installed.

      During instantiation, the value of this variable is obtained from the
      instance variable ``fg_directory`` of
      :class:`HS_Camconfig <HWS.HS_CamConfig>`.

      :Type: *string*

    - ``camera_config_file``: the name of the camera configuration file
      (e.g., dalsa_1m60.cfg)

      :Type: *string*

    - ``camera_parameters``: an instance of HS_CamConfig that holds the
      camera paremeters retrieved by the frame grabber ``gcp`` command.

      :Type: an instance of :class:`HS_Camconfig <HWS.HS_CamConfig>`
      :Updated by: :func:`~HS_Camera.get_camera_parameters`

    - ``valid_image_shape``: the horizontal and vertical dimensions of
      the CCD array of the camera.

      :Type: *tuple* of two positive *integers*

    - ``dark_image``: an image acquired with neither ambient light nor a light
      source.

      It is up to a user to ensure that the image is really dark.

      :Type: *ndarray*

    - ``frame_folder``: the folder within which the acquired image frames are
      stored.

      :Type: *string*

    - ``temperature``: a list of the camera digitizer and sensor temperatures.

      The variable has two elements, ``sensor`` and ``digitizer`` temperatures
      that one can retrieved by using the camera command ``vt``.

      :Type: *dict* of two *float* values
      :Updated by: :func:`~HS_Camera.get_temperature`

    - ``exposure_time``: the camera exposure time in seconds.

      :Type: positive *float*
      :Updated by: :func:`~HS_Camera.get_exposure_time`

    - ``timeouts``: the number of timeouts that occurred during the frame
      acquisition.

      :Type: non-negative *integer*
      :Updated by: :func:`~HS_Camera.parse_take_message`

    - ``overruns`` the number of overruns that occurred during the frame
      acquisition.

      :Type: non-negative *integer*
      :Updated by: :func:`~HS_Camera.parse_take_message`

    - ``frame_rate``: the frame rate used to acquire the frames.

      :Type: positive *integer*
      :Updated by: :func:`~HS_Camera.parse_take_message`

    - ``gpstime``: the gps time at which the most recent frame was acquired.

      :Type: positive *integer*
      :Updated by: :func:`~HS_Camera.take_frames`

    - ``bad_pixels_data``: *dict* of data gathered when the bad pixels are
      located using a dark image.

      The variable has the following fields:

      - ``csn``: the camera serial number
      - ``pv_threshold``: threshold pixel value used to determine bad pixels
      - ``coords``: the coordinates of the bad pixels (pixel units)
      - ``gpstime``: the gps time at which the last frame was taken
      - ``exposure_time``: the exposure time used to take the frames

      :Type: *dict*
      :Updated by: :func:`~HS_Camera.take_and_locate_bad_pixels`

    - ``camera_folder``: the folder in which the bad pixels data, the acquired
      frames, reference centroids, etc. are stored.

      If it is not defined by a user at instantiation,
      :func:`~HS_Camera.set_camera_folder` will set it to the default value and
      ensure that the camera folder exists.

      :Type: *string*
      :Updated by: (optional) :func:`~HS_Camera.set_camera_folder`

    - ``take_command_list``: a list of a command and parameters used to take the
      image frames.

      :Type: *list* of *strings*
      :Updated by: :func:`~HS_Camera.build_take_command`

      It is not used by the instance methods, but is here so that a user can
      verify what commands were used to take the image frames.

    - ``take_command_string``: a string of the command and parameters used to
      take the image frames.

      It is costructed by inserting space in between and concatenating all the
      entries in ``take_command_list``.

      :Type: *string*
      :Updated by: :func:`~HS_Camera.build_take_command`

      It is not used by the instance methods, but is here so that a user can
      verify what commands were used to take the image frames.

      .. note:: The variables below are currently not in use, and may be removed
                in future revisions unless found useful.

    - ``lever_arm``: the length of the lever arm of the camera); the distance
      between the Hartmann plate and the CCD (in meters).

      :Type: positive *float*

    - ``exposure_default_incr``
    - ``pcmax``
    - ``upper_th``
    - ``lower_th``
    - ``pixel_size``
    - ``background_std``
    - ``test_image``
    - ``responsivity``
    - ``no_of_images``
    - ``max_PV``
    - ``background``
    """

    # This is the top folder in which the camera folder will be made if you use
    # set_camera_folder method.
    master_folder = "/home/controls/camera_folder/"
    """A class variable specifying the folder inside which all camera folders
    are to be found.

    :Type: *string*
    :Required by: :func:`~HS_Camera.set_camera_folder`
    """

    def __init__(self, device_no=0, camera_folder=None, check_status=True):
        self.fg_directory = None
        self.camera_config_file = "dalsa_1m60.cfg"
        self.camera_parameters = None
        self.lever_arm = None
        self.pixel_size = 12e-6
        self.valid_image_shape = (1024, 1024)
        self.max_PV = 4095
        self.background = None
        self.background_std = None
        self.dark_image = None
        self.test_image = None
        self.responsivity = None
        self.frame_folder = None
        self.temperature = None
        self.exposure_time = None
        self.no_of_images = None
        self.timeouts = None
        self.overruns = None
        self.frame_rate = None
        self.exposure_default_incr = None
        self.pcmax = None
        self.upper_th = None
        self.lower_th = None
        self.gpstime = None
        self.bad_pixels_data = None
        self.camera_folder = camera_folder

        if (device_no != 0) and (device_no != 1):
            raise InvalidDeviceNo

        self.device_no = int(device_no)
        self.camera_parameters = HS_CamConfig()
        self.fg_directory = self.camera_parameters.fg_directory

        self.take_command_list = None
        self.take_command_str = None

        if check_status is True:
            self.check_camera_connection()
            self.get_camera_parameters()

        if (self.camera_folder is None) and (check_status is True):
            self.set_camera_folder()

        # res = self.check_camera_connection()
        # if res != 1:
        #     pass

        # self.exposure_time = self.get_exposure_time()
        # self.get_camera_parameters()

    def set_camera_folder(self):
        """Set up the default camera folder in which to store data.

        :Updates: ``camera_folder``

        """
        csn = self.camera_parameters.camera_serial_number
        cf = HS_Camera.master_folder + csn

        if os.path.exists(HS_Camera.master_folder):
            if not os.path.exists(cf):
                print("The camera folder " + csn + " does not exist." + "Creating...")
                os.makedirs(cf)
                print("Created the camera folder: " + cf + "/")
        else:
            print(
                "The master folder "
                + HS_Camera.master_folder
                + " does not exist. "
                + "Creating..."
            )
            os.makedirs(HS_Camera.master_folder)
            print("Ceated the master folder: " + HS_Camera.master_folder)
            os.makedirs(cf)
            print("Created the camera folder: " + cf)

        self.camera_folder = cf + "/"

    def on_or_off(self):
        """Check if the camera is accessible.

        :Uses: :func:`~HS_Camera.serial_cmd`
        :Returns: 0 (inaccessible) or 1 (accessible)
        :rtype: *integer*

        """
        try:
            scout = self.serial_cmd("vt")
        except SerialCommandFailure:
            print("The camera is not accessible")
            return 0

        if "OK>" in scout["stdout"]:
            print("The camera is accessible")
            return 1
        else:
            print("The camera is not accessible")
            return 0

    def check_camera_connection(self, n=1):
        """Check the camera connection multiple times.

        If the connection fails, advises a user to check the equipment. If the
        connection cannot be made after a certain number of attempts (beyond the
        first attempt), the method raises an exception.

        If n is zero, no further attempts are made beyond the first one.

        :param n: the number of times to re-attempt connecting to the camera
        :type n: *integer*
        :Uses: :func:`~HS_Camera.on_or_off`

        """
        if n > 0:
            for ii in arange(n):
                flag = self.on_or_off()
                if flag == 1:
                    break
                else:
                    foo = input(
                        "Please check the equipments, then "
                        + "hit any key to check the connection again."
                    )
        elif n == 0:
            flag = self.on_or_off()
        else:
            raise Exception("n must be a positive integer.")
        if flag != 1:
            raise CameraInaccessible

    def load_camera_config(self, ccf=None):
        """Load the camera configuration file into the frame grabber.

        This method is not in use at the moment, since connection failure often
        requires resolution that is beyond the capabilities of this class.

        May need a different approach to make this method useful.

        :param ccf: (optional) the full path file name of the configuration
                    file
        :type ccf: *string*
        :Uses: :func:`~HS_Camera.syscom`

        """
        if ccf is None:
            ccf = self.fg_directory + "camera_config/" + self.camera_config_file

        clist = [(self.fg_directory + "initcam"), "-c", str(self.device_no), "-f", ccf]

        cout = self.syscom(clist)

    def take_frames(self, folder=None, fprefix=None, no_of_frames=1, nbuf=1):
        """Take image frames using the camera.

        :param folder: the folder in which to save the image frames
        :type folder: *string*
        :param fprefix: the file prefix for the image frame files
        :param no_of_frames: the number of frames to take
        :param nbuf: the number of ring buffers to use when taking the frames
        :type nbuf: *integer* (from 1 to 4)
        :Uses: :func:`~HS_Camera.build_take_command`, :func:`~HS_Camera.syscom`,
               :func:`~HS_Camera.parse_take_message`, :func:`~HS_Camera.utc2gps`
        :Updates: ``gpstime``, ``fprefix``, ``frame_folder``, ``frame_rate``,
                  ``timeouts``, ``overruns``

        """
        if folder is None:
            folder = self.camera_folder + "frames/"

        self.frame_folder = folder

        if fprefix is None:
            fprefix = "image"

        self.fprefix = fprefix

        self.build_take_command(folder, fprefix, no_of_frames, nbuf)

        if not os.path.exists(folder):
            print("The folder named " + folder + " does not exist. " + "Creating...")
            os.makedirs(folder)
            print("Created the folder: " + folder)

        scout = self.syscom(self.take_command_list)

        self.parse_take_message(scout["stdout"])

        if (self.timeouts + self.overruns) > 0:
            raise AcquisitionFailure(self.timeouts, self.overruns)

        # utc2gps returns a GPS time in days not seconds
        self.gpstime = self.utc2gps(datetime.utcnow()) * 86400

    def take_and_locate_bad_pixels(self, pv_threshold=150, no_of_frames=100, nbuf=1):
        """Take the image frames and locate pixels whose values are higher than
        the specified threshold value.

        The method updates the instance variable ``bad_pixels_data`` which is a
        dict of several fields; for details see the description of the variable.

        The method also saves ``bad_pixel_data`` as a file.

        :param pv_threshold: the threshold pixel value to determine bad pixels
        :type pv_threshold: *float*
        :param no_of_frames: the number of image frames to take
        :type no_of_frames: positive *integer*
        :param nbuf: the number of ring buffers to use when taking the frames
        :type nbuf: *integer* (from 1 to 4)
        :Requires: ``camera_parameters.camera_serial_number``, ``camera_folder``,
                   :class:`HS_Image <HWS.HS_Image.HS_Image>`
        :Uses: :func:`~HS_Camera.take_and_average_frames`,
               :func:`HS_Image.locate_bad_pixels() <HWS.HS_Image.HS_Image.locate_bad_pixels>`,
        :Updates: ``bad_pixels_data``

        """
        csn = self.camera_parameters.camera_serial_number
        folder = self.camera_folder + "dark_images/"
        nof = no_of_frames
        nbf = nbuf
        hsi = self.take_and_average_frames(
            folder, fprefix="dark", no_of_frames=nof, nbuf=nbf
        )
        hsi.locate_bad_pixels(pv_threshold)

        bad_pixels_data = {}
        bad_pixels_data["csn"] = csn
        bad_pixels_data["pv_threshold"] = pv_threshold
        bad_pixels_data["coords"] = hsi.bad_pixels
        bad_pixels_data["gpstime"] = self.gpstime

        self.get_exposure_time()
        bad_pixels_data["exposure_time"] = self.exposure_time

        bp_fileloc = self.camera_folder + "bad_pixels_data_" + csn
        save(bp_fileloc, bad_pixels_data)

        self.bad_pixels_data = bad_pixels_data

        print(
            "Saved bad pixels data as a file "
            + "bad_pixels_data_"
            + csn
            + ".npy "
            + "in the folder "
            + self.camera_folder
        )
        return hsi

    def build_take_command(self, folder, fprefix, no_of_frames, nbuf):
        """Build a command string used to take image frames.

        :param folder: the folder in which to save the image frames
        :type folder: *string*
        :param fprefix: the file prefix for the image frame files
        :type fprefix: *string*
        :param no_of_frames: the number of frames to take
        :type no_of_frames: positive *integer*
        :param nbuf: the number of ring buffers to use when taking the frames
        :type nbuf: *integer* (from 1 to 4)
        :Requires: ``fg_directory``, ``device_no``
        :Updates: ``take_command_list``, ``take_command_string``

        """
        clist = [
            self.fg_directory + "take",
            "-c",
            str(self.device_no),
            "-N",
            str(nbuf),
            "-s",
            str(1),
            "-l",
            str(no_of_frames),
            "-f",
            folder + fprefix,
        ]

        self.take_command_list = clist
        self.take_command_str = " ".join(clist)

    def take_and_average_frames(
        self, folder=None, fprefix=None, no_of_frames=1, nbuf=1
    ):
        """Take and average the image frames, the return an instance of the
        :class:`HS_Image <HWS.HS_Image>` class.

        :param folder: the folder in which to save the image frames
        :type folder: *string*
        :param fprefix: the file prefix for the image frame files
        :type fprefix: *string*
        :param no_of_frames: the number of frames to take
        :type no_of_frames: positive *integer*
        :param nbuf: the number of ring buffers to use when taking the frames
        :type nbuf: *integer* (from 1 to 4)
        :Uses: :func:`~HS_Camera.take_frames`,
               :class:`HS_Image <HWS.HS_Image.HS_Image>`,
               :func:`HS_Image.read_image() <HWS.HS_Image.HS_Image.read_image>`
        :returns: an instance of :class:`HS_Image <HWS.HS_Image.HS_Image>`
        :rtype: :class:`HS_Image <HWS.HS_Image.HS_Image>`

        """
        self.take_frames(folder, fprefix, no_of_frames, nbuf)

        if fprefix is not None:
            self.fprefix = fprefix

        if folder is not None:
            self.frame_folder = folder

        ifolder = self.frame_folder
        ifprefix = self.fprefix
        hsi = HS_Image(
            location_type="folder",
            location=ifolder,
            fprefix=ifprefix,
            first_frame=1,
            no_of_frames=no_of_frames,
        )

        hsi.read_image()
        return hsi

    def parse_take_message(self, scstr):
        """Parse the output message from the frame grabber to retrieve
        ``frame_rate``, ``timeouts``, and ``overruns`` values.

        :param scstr: the output message
        :type scstr: *string*
        :Updates: ``frame_rate``, ``timeouts``, ``overruns``

        """
        pstr = scstr.split("\n")

        for fs in pstr:
            if "frames/sec" in fs:
                self.frame_rate = float(fs.split(" ")[0])

            if "overruns" in fs:
                stlist = fs.strip().split(" ")

        self.timeouts = int(stlist[2])
        self.overruns = int(stlist[4])

    def check_pixel_maximum(self, im):
        """(Not implemented Yet)"""
        # if im.max() > self.max_PV:
        pass

    def check_image_shape(self, im):
        """Check the shape of the image array.

        The method is not useful at the moment as the class already assumes that an
        image frame is of certain shape when reading it.

        """
        if shape(im) != self.valid_image_shape:
            raise InvalidImageShape(shape(im), self.valid_image_shape)

    def get_camera_parameters(self):
        """Get camera parameters.

        :Uses: :class:`HS_CamConfig <HWS.HS_CamConfig.HS_CamConfig>`,
               :func:`HS_CamConfig.get_parameters() <HWS.HS_CamConfig.HS_CamConfig.get_parameters>`,
               :func:`~HS_Camera.get_temperature`
        :Updates: ``camera_parameters``

        """
        self.camera_parameters = HS_CamConfig()
        self.camera_parameters.get_parameters(self)
        self.get_temperature()

    def get_temperature(self):
        """Get the sensor and digitizer temperatures of the camera

        :Uses: :func:`~HS_Camera.serial_cmd`
        :Updates: ``temperature``

        """
        temperature = {}

        scout = self.serial_cmd("vt")
        pstr = []
        if "OK>" in scout["stdout"]:
            vtstr = scout["stdout"]
            for fs in vtstr.split("\n"):
                fsl = fs.split(":")
                if (size(fsl) == 2) and (len(fsl[1]) > 0):
                    pstr = pstr + [fsl[1].strip()]
        else:
            print("Warning: Failed to obtain camera temperature values.")

        temperature["digitizer"] = float(pstr[0].split(" ")[0])
        temperature["sensor"] = float(pstr[1].split(" ")[0])

        self.temperature = temperature

    def get_exposure_time(self):
        """Get the exposure time (in units of seconds) used to take the image frames.

        :Uses: :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``exposure_time``

        """
        self.get_camera_parameters()
        et = self.camera_parameters.exposure_time.split(" ")[0]
        et = float(et) * 1e-6
        self.exposure_time = et

    def serial_cmd(self, sc):
        """Run a camera serial command.

        :param sc: the serial command to be run
        :type sc: *string* or *list* of *strings*
        :Requires: ``fg_directory``, ``device_no``
        :Uses: :func:`~HS_Camera.syscom`
        :Returns: the output message generated by running the command.
        :rtype: *dict*

        """
        fgdir = self.fg_directory
        cmdlist = [(fgdir + "serial_cmd"), "-c", str(self.device_no)]

        cmdstr = (fgdir + "serial_cmd") + " -c " + str(self.device_no)
        scstr = cmdstr + " " + sc

        if type(sc) is str:
            sclist = cmdlist + [sc]
        elif type(sc) is list:
            sclist = cmdlist + sc
        else:
            raise InvalidCommandFormat

        scout = self.syscom(sclist)

        if scout["stdout"] == "\n":
            # return scout
            raise SerialCommandFailure(sc, scstr)
        elif "Error" in scout["stdout"]:
            # return scout
            raise SerialCommandError(sc, scstr, scout)

        return scout

    def serial_cmd_alt(self, sc):
        """Run a camera serial command.

        Unlike :func:`~HS_Camera.serial_cmd`, this method passes a serial command
        and its associated parameters as a string to :func:`~HS_Camera.syscom_alt`,
        which is necessary if the serial command to be run has extra parameters.

        The serial command ``sc`` must be a string encapsulated by the double
        quotation marks.

        :param sc: the serial command to be run
        :type sc: *string*
        :Requires: ``fg_directory``, ``device_no``
        :Uses: :func:`~HS_Camera.syscom_alt`
        :Returns: the output message generated by running the command.
        :rtype: *dict*

        """
        fgdir = self.fg_directory
        cmdstr = (fgdir + "serial_cmd") + " -c " + str(self.device_no)
        scstr = cmdstr + " " + sc

        scout = self.syscom_alt(scstr)

        if scout["stdout"] == "\n":
            return scout
            raise SerialCommandFailure(sc, scstr)
        elif "Error" in scout["stdout"]:
            # return scout
            raise SerialCommandError(sc, scstr, scout)

        return scout

    def syscom(self, sclist):
        """Run a system command.

        :param sclist: a *list* of the command to be run and the parameters
        :type sclist: *list*
        :Returns: a *dict* consisting of returncode, stdout, and stderr output
        :rtype: *dict*

        """
        mycom = sp.Popen(sclist, stdout=sp.PIPE, stderr=sp.PIPE)
        sco = mycom.communicate()
        scout = {}
        scout["returncode"] = mycom.returncode
        scout["stdout"] = sco[0].decode()
        scout["stderr"] = sco[1].decode()

        return scout

    def syscom_alt(self, scstr):
        """Run a system command.

        Unlike :func:`~HS_Camera.syscom`, this method passes the string of a command
        and its associated parameters as a *string* instead of as a *list*. This is
        necessary when running a camera serial command that contains extra parameters.

        Other system commands with multiple parameters can be run normally using
        :func:`~HS_Camera.syscom`, so this method is used only for some camera serial
        commands.

        :param scstr: a *string* of a system command and parameters
        :type scstr: *string*
        :Returns: a *dict* consisting of returncode, stdout, and stderr output
        :rtype: *dict*

        """
        mycom = sp.Popen(scstr, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        sco = mycom.communicate()
        scout = {}
        scout["returncode"] = mycom.returncode
        scout["stdout"] = sco[0].decode()
        scout["stderr"] = sco[1].decode()

        return scout

    def set_output_mode(self, mode, to_output=False):
        """Set the output mode for the camera.

        ``mode`` takes one of two values:
        - 0: single tap readout, maximum 30 fps
        - 1: dual tap readout, maximum 60 fps (default)

        :param mode: 0 or 1
        :type mode: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"sos ' + str(mode) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_data_mode(self, mode, to_output=False):
        """Set the data mode for the camera.

        ``mode`` takes one of three values:
        - 12: 12-bit (default)
        - 10: 10-bit
        - 8: 8-bit

        :param mode: 8, 10 or 12
        :type mode: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"oms ' + str(mode) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_baud_rate(self, rate, to_output=False):
        """Set the serial communication speed for the camera.

        ``rate`` specifies the baud rate of the communication and can take one of
        three values: 9600 (default), 19200 or 57600.

        :param rate: 9600, 19200 or 57600
        :type rate: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"sbr ' + str(rate) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_exposure_mode(self, mode, to_output=False):
        """Set the exposure mode of the camera.

        See the camera manual Section 3.7 for a detailed explanation on the exposure
        mode, exposure time, and the frame rate.

        :param mode: *integer* from 2 to 7
        :type mode: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"sem ' + str(mode) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_exposure_control(self, mode, to_output=False):
        """Enable or disable the camera exposure control.

        ``mode`` takes one of two values:

        - 0: disable exposure control
        - 1: enable exposure control

        Relevant only if the camera is in the exposure mode 3

        :param mode: 0 or 1
        :type mode: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"sec ' + str(mode) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_frame_rate(self, rate, to_output=False):
        """Set the frame acquisition rate.

        See section 3.7.3 of the camera manual for the details.

        :param rate: the frame acquisition rate
        :type rate: positive *float*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"ssf ' + str(rate) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_exposure_time(self, duration, to_output=False):
        """Set the exposure time (in microseconds) used to acquire a frame.

        See section 3.7.3 of the camera manual for the details.

        :param duration: the exposure time duration
        :type duration: positive *float*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        self.set_exposure_mode(6)
        scout = self.serial_cmd_alt('"set ' + str(duration) + '"')
        self.set_exposure_mode(2)
        scout = self.get_camera_parameters()
        et = self.get_exposure_time()

        if to_output == True:
            return scout

    def set_analog_gain(self, tap, gain, to_output=False):
        """Set the analog gain of the camera.

        See section 3.8 of the camera manual for the details.

        :param tap: tap value of 1 or 2
        :type tap: *integer*
        :param gain: gain value between 0 and 4095
        :type gain: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"ssg ' + str(tap) + " " + str(duration) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_digital_gain(self, gain, to_output=False):
        """Set the digital gain of the camera.

        See section 3.8 of the camera manual for the details.

        :param gain: a value from 0 (default) to 2
        :type gain: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"gm ' + str(gain) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_binning_mode(self, hor, ver, to_output=False):
        """Set the binning mode of the camera.

        See section 3.9 of the camera manual for the details.

        Both parameters take a value from 1, 2, 4 and 8.

        :param hor: horizontal binning value (1, 2, 4 or 8)
        :type hor: *integer*
        :param ver: vertical binning value (1, 2, 4 or 8)
        :type ver: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"ssg ' + str(hor) + " " + str(ver) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def reboot_camera(self, to_output=False):
        """Reboots the camera.

        Rebooting the camera restores its setting to the previously saved one.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("rc")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_pretrigger(self, pt, to_output=False):
        """Set the camera pre-trigger value.

        May be relevant for some frame grabbers.

        :param pt: a value from 0 to 15
        :type pt: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"sp ' + str(pt) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def set_video_mode(self, mode, to_output=False):
        """Set the video mode of the camera.

        The mode 0 is the normal operating mode.

        Setting the mode to a value other than 0 displays the corresponding test
        pattern image.

        See section 3.12 of the camera manual for the details.

        :param mode: a value from 0 to 11
        :type mode: *integer*
        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd_alt`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd_alt('"svm ' + str(mode) + '"')
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def get_camera_serial_no(self, to_output=True):
        """Get the output that contains the camera serial number.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("gcs")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def get_camera_model(self, to_output=True):
        """Get the output that contains the camera model number.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("gcm")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def get_camera_version(self, to_output=True):
        """Get the output that contains the camera version.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("gcv")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def get_sensor_serial_no(self, to_output=True):
        """Get the output that contains the camera version.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("gss")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def restore_factory_setting(self, to_output=False):
        """Restore the camera setting to the factory setting.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("rfs")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def restore_user_setting(self, to_output=False):
        """Restore the camera setting to the last saved user setting.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("rus")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def write_user_setting(self, to_output=False):
        """Save the current camera setting.

        :param to_output: ``True`` to return the output list, ``False``
                          to run quietly
        :type to_output: *boolean*
        :Uses: :func:`~HS_Camera.serial_cmd`,
               :func:`~HS_Camera.get_camera_parameters`
        :Updates: ``camera_parameters``
        :Returns: (optional) a list of the output from running the command
        :rtype: *dict*

        """
        scout = self.serial_cmd("wus")
        self.get_camera_parameters()

        if to_output == True:
            return scout

    def utc2gps(self, mydt):
        """Convert utc time to gps time.

        The input parameter ``mydt`` is generated by running ``utcnow()`` method
        of the ``datetime`` module. It is an instance of ``datetime`` class.

        This method then converts the time into gps time, taking into account of
        the stepdates.

        The list of the stepdates ``stepdates_t`` should be revised as we get
        more step dates in the future.

        :param mydt: utc datetime
        :type mydt: *datetime.datetime*
        :Returns: the converted gps time
        :rtype: *float*

        """
        # A date for conversion should be later than first_valid_date.
        first_valid_date = (1980, 1, 6)

        # The dates on which the leap seconds were announced.
        stepdates_t = [
            (1981, 7, 1),
            (1982, 7, 1),
            (1983, 7, 1),
            (1985, 7, 1),
            (1988, 1, 1),
            (1990, 1, 1),
            (1991, 1, 1),
            (1992, 7, 1),
            (1993, 7, 1),
            (1994, 7, 1),
            (1996, 7, 1),
            (1997, 7, 1),
            (1999, 1, 1),
            (2006, 1, 1),
            (2009, 1, 1),
            (2012, 7, 1),
            (2015, 7, 1),
            (2017, 1, 1),
        ]

        stepdates = [datetime(*sdt) for sdt in stepdates_t]

        # Determine how many leap seconds should be added
        leapseconds = sum([mydt > sdt for sdt in stepdates])

        # Time difference (in seconds) between the given date and the start date
        td = mydt - datetime(*first_valid_date)

        # return the time difference (with leap seconds added) in units of days
        return (td.total_seconds() + leapseconds) / (24 * 3600)


class InvalidDeviceNo(Exception):
    def __init__(self):
        self.message = "device_no must be 0 or 1."

    def __str__(self):
        return self.message


class CameraInaccessible(Exception):
    def __init__(self):
        self.message = "Unable to connect to the camera."

    def __str__(self):
        return self.message


class AcquisitionFailure(Exception):
    def __init__(self, timeouts, overruns):
        self.timeouts = timeouts
        self.overruns = overruns
        self.message = (
            "Image acquisition failed: "
            + "{} timeouts and ".format(timeouts)
            + "{} overruns occurred.".format(overruns)
        )

    def __str__(self):
        return self.message


class InvalidImageShape(Exception):
    def __init__(self, image_shape, valid_image_shape):
        self.image_shape = image_shape
        self.valid_image_shape = valid_image_shape
        self.message = (
            "Image shape ({0[0]},{0[1]}) is not valid: " + "Expected ({1[0]},{1[1]})."
        ).format(image_shape, valid_image_shape)

    def __str__(self):
        return self.message


class InvalidCommandFormat(Exception):
    def __init__(self):
        self.message = (
            "The camera serial command arguments must be "
            + "a string or a list of strings."
        )

    def __str__(self):
        return self.message


class SerialCommandFailure(Exception):
    def __init__(self, command, cstr):
        self.command = command
        self.command_string = cstr
        if cstr is None:
            self.message = "Failed to execute the serial command."
        else:
            self.message = "Failed to execute the serial command: " + command

    def __str__(self):
        return self.message


class SerialCommandError(Exception):
    def __init__(self, command, cstr, scout):
        self.command = command
        self.command_string = cstr
        self.command_output = scout
        self.message = (
            "Error occurred when executing the serial command: "
            + command
            + "\n\n"
            + "The error message is as follows:\n"
            + scout["stdout"]
        )

    def __str__(self):
        return self.message


# Won Kim 23/08/2018 12:47:41 PM



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from HWS.HS_Camera import *
import os


class HS_CamConfig:
    """A class to hold the camera configuration data.

    ``HS_CamConfig`` is a class that retrieves and stores the camera configuration
    of Dalsa Pantera 1M60 camera.

    The class needs :class:`HS_Camera <HWS.HS_Camera.HS_Camera>` to retrieve the
    camera configuration, and will not work by itself. (May change in a future)

    **Instantiation**

    Although the class can be instantiated separately, a user will not have to
    do so in most cases. When :class:`HS_Camera <HWS.HS_Camera.HS_Camera>` is
    instantiated, ``HS_CamConfig`` is also instantiated and assigned as one of
    its instance variables.

    **Instance Variables**

    The instance variables consist of those which are retrieved when the serial
    command 'gcp' is executed. The sensor and digitizer temperature values are
    not retrieved by 'gcp' but another command, and is assigned to an instance
    variable of :class:`HS_Camera <HWS.HS_Camera.HS_Camera>`.

    For descriptions of each camera parameters, consult the camera manual.

    """

    def __init__(self):
        # self.fg_directory = '/opt/EDTpdv/'
        self.fg_directory = os.getenv("FG_DIRECTORY")
        self.model_number = None
        self.camera_serial_number = None
        self.sensor_serial_number = None
        self.tap_1_gain = None
        self.tap_2_gain = None
        self.firmware_design_rev = None
        self.dsp_design_rev = None
        self.pretrigger = None
        self.video_mode = None
        self.data_mode = None
        self.binning_mode = None
        self.gain_mode = None
        self.output_configuration = None
        self.exposure_control = None
        self.exposure_mode = None
        self.sync_frequency = None
        self.exposure_time = None

    def get_parameters(self, cam):
        """Get camera configuration parameters.

        :param cam: an instance of :class:`HS_Camera <HWS.HS_Camera.HS_Camera>`
        :Uses: :func:`HS_Camera.serial_cmd() <HWS.HS_Camera.HS_Camera.serial_cmd>`,
               :func:`~HS_CamConfig.parse_ccd_output`

        """
        gcpout = cam.serial_cmd("gcp")
        # Dan+Cao : Unsure why this has been commented out
        # if 'OK>' not in gcpout['stdout']:
        #     raise \
        #       Exception('Failed to retrieve the parameters from the camera.')
        self.parse_ccd_output(gcpout["stdout"])

    def parse_ccd_output(self, gcpstr):
        """Parse the output from 'gcp' command, and updates the camera parameters.

        :param gcpstr: an output from 'gcp' command
        :type gcpstr: *string*
        :Updates: all instance variables of ``HS_CamConfig``

        """
        pstr = []

        for fs in gcpstr.split("\n"):
            fsl = fs.split(":")
            if (size(fsl) == 2) and (len(fsl[1]) > 0):
                pstr = pstr + [fsl[1].strip()]

        self.model_number = pstr[0]
        self.camera_serial_number = pstr[1]
        self.sensor_serial_number = pstr[2]
        self.tap_1_gain = pstr[3]
        self.tap_2_gain = pstr[4]
        self.firmware_design_rev = pstr[5]
        self.dsp_design_rev = pstr[6]
        self.pretrigger = pstr[7]
        self.video_mode = pstr[8]
        self.data_mode = pstr[9]
        self.binning_mode = pstr[10]
        self.gain_mode = pstr[11]
        self.output_configuration = pstr[12]
        self.exposure_control = pstr[13]
        self.exposure_mode = pstr[14]
        self.sync_frequency = pstr[15]
        self.exposure_time = pstr[16]



"""
This file is *not* licensed under the Unviersity of Adelaide license as laid out in
the accompanying LICENSE file.

HS_Export.py exports data to file

4-Oct-2018: updated the HS_Export class to export to HDF5 instead of pickle. The file content is the same but the export format is now HDF5
          : Also set the avImage to be compressed. This looks to reduce the file size by slightly more than 50%
          : 'datetime' does not transfer easily to HDF5 so this is converted to ISOFORMAT
"""
import os
import subprocess as sp
from socket import gethostname
from numpy import *
from datetime import datetime, timedelta
import pickle
from time import strftime, gmtime
import h5py

from HWS.HS_CamConfig import *
from HWS.HS_Image import *


class HS_Export:
    # A class to export data to file.

    def __init__(self, IFO="H1", optic="ETMY_HWS", frameDir="/home/controls/temp"):
        self.avGradients = None
        self.avGpstime = 0
        self.avImage = None
        self.numberOfAvg = 0
        self.lastSaveTime = datetime.now()
        self.averageDuration = 20
        self.IFO = IFO
        self.optic = optic
        self.frameDir = frameDir
        self.timeDir = None
        self.filename = None
        self.hostname = gethostname()
        self.origin_x = None
        self.origin_y = None
        self.mask_radius = None
        self.mask_center_x = None
        self.mask_center_y = None

    def update_averages(self, gradients, image, gpstime):
        # update the averages to be saved
        if self.numberOfAvg == 0:
            self.avGradients = gradients
            self.avGpstime = gpstime
            self.avImage = image

            self.numberOfAvg = 1
        else:
            self.avGradients = self.avGradients + gradients
            self.avGpstime = self.avGpstime + gpstime
            self.avImage = self.avImage + image

            self.numberOfAvg = self.numberOfAvg + 1

    def export_if_necessary(self):
        # export the data to file if enough time has elapsed
        tdelta = datetime.now() - self.lastSaveTime

        if self.numberOfAvg > 0 and (tdelta.total_seconds() > self.averageDuration):
            # get the averages
            self.avGradients = self.avGradients / self.numberOfAvg
            self.avGpstime = self.avGpstime / self.numberOfAvg
            self.avImage = self.avImage / self.numberOfAvg
            self.avImage = self.avImage.astype("uint16")

            # udpate the save time so it is stored in the file
            self.lastSaveTime = datetime.now()

            # create the time directory string and check it exists
            self.timeDir = str(int(floor(self.avGpstime / 1e5) * 1e5)) + "/"
            dirSub = (
                self.frameDir + "/" + self.IFO + "/" + self.optic + "/" + self.timeDir
            )

            if os.path.exists(dirSub) == False:
                # create the sub directory
                os.makedirs(dirSub)

            # create the filename and export the data
            self.filename = (
                dirSub
                + str(int(round(self.avGpstime)))
                + "_"
                + self.IFO
                + "_"
                + self.optic
                + "_HS_Export.hdf5"
            )
            # pickle.dump(self,open(self.filename,"wb"),2)
            print(self.filename)

            # export the data using HDF5
            f = h5py.File(self.filename, "w")
            f["avGradients"] = self.avGradients
            f["avGradients"].attrs["numberOfAvg"] = self.numberOfAvg
            f["avGradients"].attrs["lastSaveTime"] = self.lastSaveTime.isoformat()
            f["avGradients"].attrs["avGpstime"] = self.avGpstime
            f["avGradients"].attrs["averageDuration"] = self.averageDuration
            f["avGradients"].attrs["IFO"] = self.IFO
            f["avGradients"].attrs["optic"] = self.optic
            f["avGradients"].attrs["frameDir"] = self.frameDir
            f["avGradients"].attrs["timeDir"] = self.timeDir
            f["avGradients"].attrs["filename"] = self.filename
            f["avGradients"].attrs["hostname"] = self.hostname
            f["avGradients"].attrs["mask_radius"] = self.mask_radius
            f["avGradients"].attrs["mask_center_x"] = self.mask_center_x
            f["avGradients"].attrs["mask_center_y"] = self.mask_center_y
            f["avGradients"].attrs["origin_x"] = self.origin_x
            f["avGradients"].attrs["origin_y"] = self.origin_y

            dset = f.create_dataset("avImage", data=self.avImage, compression="gzip")
            f.close()

            print(
                "Saved file - "
                + strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
                + " UTC"
            )

            # reset the averages
            self.numberOfAvg = 0



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *


class HS_GaussianWF:
    def __init__(self):
        amplitude = None
        width = None
        gradients = None
        wf = None
        wf_full = None
        centroids = None
        centre = array([511.5, 511.5]) * 12e-6

    def synthesize_gradients(self):
        ctr = zeros(shape(self.centroids))
        ctr[:, 0] = self.centroids[:, 0] - self.centre[0]
        ctr[:, 1] = self.centroids[:, 1] - self.centre[1]

        self.construct_wf()

        self.gradients = zeros(shape(ctr))
        self.gradients[:, 0] = self.wf * (2 / self.width ** 2) * ctr[:, 0]
        self.gradients[:, 1] = self.wf * (2 / self.width ** 2) * ctr[:, 1]

    def construct_wf(self):
        ctr = zeros(shape(self.centroids))
        ctr[:, 0] = self.centroids[:, 0] - self.centre[0]
        ctr[:, 1] = self.centroids[:, 1] - self.centre[1]

        self.wf = self.amplitude * exp(
            -(ctr[:, 0] ** 2 + ctr[:, 1] ** 2) / self.width ** 2
        )



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *


class HS_GaussianBeam:
    def __init__(
        self,
        wavelength=None,
        distance_from_waist=None,
        waist_radius=None,
        max_amplitude=None,
        grid_size=None,
        grid_spacing=None,
    ):

        self.wavelength = wavelength
        self.distance_from_waist = distance_from_waist
        self.waist_radius = waist_radius
        self.max_amplitude = max_amplitude
        self.grid_size = grid_size
        self.grid_spacing = grid_spacing

        self.grid_x = None
        self.grid_y = None
        self.guoy_phase = 0
        self.curvature = None
        self.radius = None
        self.amplitude = None
        self.intensity = None
        self.rayleigh_range = None
        self.depth_of_focus = None

        # self.__dict__.update(parameters)
        self.derive_variables()

    def derive_variables(self):
        self.rayleigh_range = pi * self.waist_radius ** 2 / float(self.wavelength)
        self.depth_of_focus = 2.0 * self.rayleigh_range
        if self.distance_from_waist == 0:
            self.curvature = inf
        else:
            self.curvature = self.distance_from_waist * (
                1 + (self.rayleigh_range / float(self.distance_from_waist)) ** 2
            )
        self.radius = (
            self.waist_radius
            * (1 + (self.distance_from_waist / float(self.rayleigh_range)) ** 2) ** 0.5
        )
        self.setup_grid()
        self.construct_amplitude()

    def setup_grid(self):

        origin = self.grid_size / 2.0 - 0.5
        xr = (arange(0, self.grid_size) - origin) * self.grid_spacing
        yr = (arange(0, self.grid_size) - origin) * self.grid_spacing

        self.grid_x, self.grid_y = meshgrid(xr, yr)

    def construct_amplitude(self):
        i = 1j
        E_0 = float(self.max_amplitude)
        w_0 = float(self.waist_radius)
        w = float(self.radius)
        r = (self.grid_x ** 2 + self.grid_y ** 2) ** 0.5
        rc = float(self.curvature)
        gph = self.guoy_phase
        z = float(self.distance_from_waist)
        k = 2 * pi / self.wavelength
        self.amplitude = (
            E_0
            * (w_0 / w)
            * exp(-((r / w) ** 2) - i * k * z - i * k * (r ** 2 / (2 * rc)) + i * gph)
        )
        self.intensity = abs(self.amplitude) ** 2

    def construct_amplitude_at_waist(self):
        E_0 = float(self.max_amplitude)
        w = self.waist_radius
        r = (self.grid_x ** 2 + self.grid_y ** 2) ** 0.5
        self.amplitude = E_0 * exp(-((r / w) ** 2))
        self.intensity = abs(self.amplitude) ** 2

    def calculate_gradients(self, centroids):
        normal_v = self.calculate_normal_vectors(centroids)
        if self.distance_from_waist > 0:
            gradients = (normal_v - array([0, 0, 1]))[:, 0:2]
        else:
            gradients = (array([0, 0, 1]) - normal_v)[:, 0:2]
        return gradients

    def calculate_normal_vectors(self, centroids):
        sc = shape(centroids)[0]
        normal_vectors = zeros([sc, 3])
        centroids = centroids.astype(float)

        for k in arange(sc):
            nv = zeros(3)
            nv[0] = centroids[k, 0]
            nv[1] = centroids[k, 1]
            nv[2] = (
                self.curvature ** 2 - (centroids[k, 0] ** 2 + centroids[k, 1] ** 2)
            ) ** 0.5

            nv = nv / nv[2]
            normal_vectors[k, :] = nv

        return normal_vectors



from . import HS_Camera
from . import HS_Centroids
from . import HS_Gradients
from . import HS_Image
from . import HS_WFP

# Set the Finesse.ligo version.
try:
    from ._version import version as __version__
except ImportError:
    raise Exception("Could not find _version.py. Ensure you have run setup.")



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *
from HWS.HS_Centroids import *
from HWS.HS_Gradients import *
from HWS.HS_WFP import *


class HS_Anomalies:
    """A class to analyse the gradients for anomalies.

    (Experimental)

    This class can be used to detect the anomalous gradient vectors that
    may appear due to undesired and unexpected small size bumps (whose
    dimensions being comparable to the spot spacing) of high magnitude in a
    transverse wavefront aberration.

    The class is instantiated with an instance of
    :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`, and analyses
    the gradients by calculating the average of the radial components of the
    gradient vectors on nearest neighbor centroids for each centroid.

    The analysis is performed on "difference gradients", the gradients from
    reference and live centroids minus the gradients from LSQ polynomial
    calculation.

    If an averaged radial component on a centroid deviates from the overall
    average by a threshold factor times the standard deviation of all radial
    magnitudes, we assume that there is an anomalous aberration near that
    centroid.

    After identifying such "anomalous" centroids, the class analyses if such
    centroids are next to each other, and if so groups them. The number of
    groups obtained as such are the number of anomalies that may be present in
    the transverse aberration. The centroid with the highest radial magnitude in
    a group is considered to be the "centre" of that group.

    The class then estimates the position of each anomaly by performing a "centre
    of mass calculation" about the centre of each group, treating the averaged
    radial gradient magnitudes of neighbouring centroids like a mass. The
    estimate is by no means supposed to be quantitatively accurate but may still
    be useful analysing a large set of data for anomalies.

    """

    def __init__(self, hsg=None, std_threshold=5, separation=40):
        # separation is a nominal separation between neighboring spots in
        # an image (i.e., spot spacing). It is used to find the nearest
        # neighbor centroids and does not need to be too accurate.
        #
        # If the value is too small, one cannot find any neighboring
        # centroids. If too big, more centroids than just the nearest
        # neighbours will be included (which sometimes one may wish to do).
        self.separation = separation
        self.hsg = hsg
        self.gradients = None
        self.difference_gradients = None
        self.reference_centroids = None
        self.gradient_magnitudes = None
        self.wfp_whole = None
        self.wfp_diff = None
        self.wf_numerical_diff = None
        self.radial_magnitudes = None
        self.radial_ratios = None
        self.radial_mean = None
        self.radial_std = None
        self.std_threshold = std_threshold
        self.indices = None
        self.groups = {}
        self.centroids = None
        self.magnitudes = None
        self.ratios = None

        self.initialise_groups()

        if (self.hsg is not None) and self.hsg.__class__.__name__ == "HS_Gradients":
            self.reference_centroids = self.hsg.initial_centroids.centroids
            self.gradients = self.hsg.gradients[:, 0:2]
            self.gradient_magnitudes = self.hsg.magnitudes
            self.analyse_gradients()

    def initialise_groups(self):
        self.groups["indices"] = []
        self.groups["radial_magnitudes"] = []
        self.groups["max_indices"] = []

    def analyse_gradients(self):
        self.construct_difference_gradients()
        self.compute_radial_gradients()
        if size(self.indices) != 0:
            self.determine_anomalous_groups()

    def construct_difference_gradients(self):
        hsg = self.hsg

        self.wfp_whole = HS_WFP(hsg)
        self.wfp_whole.compute_poly_coeffs()
        gr_BTA = self.wfp_whole.calculate_gradients()

        self.difference_gradients = self.gradients - gr_BTA

        hsc_ref = self.hsg.initial_centroids
        mod_factor = hsg.pixel_size / (hsg.lever_arm * hsg.magnification)

        hsc_live = HS_Centroids()
        hsc_live.centroids = hsc_ref.centroids + self.difference_gradients / mod_factor
        hsc_live.no_of_centroids = hsc_ref.no_of_centroids
        hsc_live.average_centroids()

        hsg_diff = HS_Gradients(hsc_ref, hsc_live)
        hsg_diff.construct_gradients()

        self.wfp_diff = HS_WFP(hsg_diff)
        self.wfp_diff.compute_poly_coeffs()

    def compute_radial_gradients(self):
        refc = self.reference_centroids.copy()
        sep = self.separation
        gr_r = []
        noc = shape(refc)[0]
        # For each coord in pos_arr, find the reference centroids that are within a
        # radius given by separation.
        for cidx in range(noc):
            # Move the centroid sufficiently far away so that it is not
            # picked by where clause below (which then will pick only the
            # neighboring centroids but not the centroid itself.
            refc[cidx, :] = refc[cidx, :] + array([sep * 10, sep * 10])
            tc = self.reference_centroids[cidx, :]
            nn_ind = where(sqrt(((refc - tc) ** 2).sum(axis=1)) < sep)
            refc[cidx, :] = self.reference_centroids[cidx, :]
            v_dis = refc[nn_ind, :] - refc[cidx, :]
            v_mag = sqrt((v_dis ** 2).sum(axis=1))
            vn_dis = v_dis / v_mag

            gr = self.difference_gradients[nn_ind, 0:2]
            gr_r.append((vn_dis * gr).mean())

        gr_r = array(gr_r)
        self.radial_magnitudes = gr_r
        self.radial_mean = gr_r.mean()
        self.radial_std = gr_r.std()

        gr_r_c = self.radial_magnitudes - self.radial_mean
        gr_r_ratio = gr_r_c / self.radial_std

        self.ratios = gr_r_ratio
        self.indices = where(abs(gr_r_ratio) > self.std_threshold)[0]

    def determine_anomalous_groups(self):
        actr = self.reference_centroids[self.indices].copy()
        groups_list = None

        # group the centroid indices
        for aind in self.indices:
            tc = self.reference_centroids[aind]
            nn_aind = where(sqrt(((actr - tc) ** 2).sum(axis=1)) < self.separation)[0]
            nn_set = set(self.indices[nn_aind])

            if groups_list is None:
                groups_list = []
                groups_list.append(nn_set)
            else:
                updated = False
                ii = 0
                for grs in groups_list:
                    if len(grs.intersection(nn_set)) != 0:
                        groups_list[ii] = grs.union(nn_set)
                        updated = True
                        break

                    ii = ii + 1

                if updated == False:
                    groups_list.append(nn_set)

        self.initialise_groups()
        for grs in groups_list:
            grs = array(list(grs))
            maxind = grs[abs(self.radial_magnitudes[grs]).argmax()]
            self.groups["max_indices"].append(maxind)
            self.groups["indices"].append(grs)
            self.groups["radial_magnitudes"].append(self.radial_magnitudes[grs])

        self.estimate_positions()
        self.estimate_positions_alt()

    def estimate_positions(self):
        """Estimate the positions of the anomalies.

        The position of each anomaly is estimated by performing the
        weighted average of the centroids, using the averaged radial
        gradient magnitudes as weights, about the centroid with the highest
        radial gradient magnitude.

        """
        self.groups["positions"] = []

        refc = self.reference_centroids
        sep = self.separation
        rm = self.radial_magnitudes

        for cidx in self.groups["max_indices"]:
            tc = self.reference_centroids[cidx]
            nn_ind = where(sqrt(((refc - tc) ** 2).sum(axis=1)) < 1.5 * sep)[0]
            pos_x = (rm[nn_ind] * refc[nn_ind, 0]).sum() / rm[nn_ind].sum()
            pos_y = (rm[nn_ind] * refc[nn_ind, 1]).sum() / rm[nn_ind].sum()

            self.groups["positions"].append([pos_x, pos_y])

        self.groups["positions"] = array(self.groups["positions"])

    def estimate_positions_alt(self):
        """Estimate the positions of the anomalies using slightly different method.

        The method is similar to :func:`~HS_Anomalies.estimate_positions`
        except that the positions are estimated by performing two weighted
        averaging successively. The first averaging is over the radial
        gradient magnitudes over the centroids that form a cluster for each
        anomalies. The second averaging is then carried out about the
        position from the first averaging as the centre for the second averaging.

        """

        self.groups["positions_alt"] = []
        self.groups["positions_from_indices"] = []

        refc = self.reference_centroids
        sep = self.separation
        rm = self.radial_magnitudes

        for g in self.groups["indices"]:
            grm = self.radial_magnitudes[g]
            ipos_x = (grm * self.reference_centroids[g, 0]).sum() / grm.sum()
            ipos_y = (grm * self.reference_centroids[g, 1]).sum() / grm.sum()

            tc = array([ipos_x, ipos_y])
            nn_ind = where(sqrt(((refc - tc) ** 2).sum(axis=1)) < 2 * sep)[0]
            pos_x = (rm[nn_ind] * refc[nn_ind, 0]).sum() / rm[nn_ind].sum()
            pos_y = (rm[nn_ind] * refc[nn_ind, 1]).sum() / rm[nn_ind].sum()

            self.groups["positions_from_indices"].append(tc)
            self.groups["positions_alt"].append([pos_x, pos_y])

        self.groups["positions_alt"] = array(self.groups["positions_alt"])


# Won Kim 5/11/2018 3:37:03 PM



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *
import scipy.ndimage

try:
    from skimage.filter import threshold_otsu
except ImportError:
    from skimage.filters import threshold_otsu

from HWS.HS_Image import *


class HS_Centroids:
    """A class to handle centroiding an image.

    ``HS_Centroids`` is a class to generate and interact with centroids
    of Hartmann Sensor images.

    **Instance Variables**

    .. note:: Some variables are here as a result of the porting process
              (from Matlab to Python), and may be removed in a future
              revision.

    - ``centroids``: an N by 2 array of the spot positions where N is the
      number of centroids.

      :Type: *ndarray*
      :Required by: :func:`~HS_Centroids.average_centroids`
      :Updated by: :func:`~HS_Centroids.find_centroids`

    - ``average``: a 1 by 2 array of the average of all centroids.

      :Type: *ndarray*
      :Updated by: :func:`~HS_Centroids.average_centroids`

    - ``hsimage``: an instance of ``HS_Image``, which contains an image
      to be centroided.

      :Type: an instance of :class:`HS_Image <HWS.HS_Image>`
      :Required by: :func:`~HS_Centroids.find_centroids_from_image`,
                    :func:`~HS_Centroids.find_centroids_using_template`

    - ``threshold``: (optional) a value specifying the threshold of an image pixel
      values.

      If it is not *None*, each pixel value of
      ``hsimage.modified_image`` that are smaller than ``threshold`` is set to
      zero before centroiding.

      :Required by: (optional) :func:`~HS_Centroids.find_centroids_using_template`

    - ``no_of_centroids``: the total number of centroids in a image.

      :Type: *integer*
      :Updated by: :func:`~HS_Centroids.find_centroids_using_template`

    - ``rejected_centroids``: (optional) the centroids rejected by the method
      :func:`~HS_Centroids.find_centroids_from_image` with the input parameter
      ``to_remove_small_spots`` set to ``True``

      :Type: *ndarray*
      :Updated by: (optional) :func:`~HS_Centroids.find_centroids_from_image`

      (WK) The term "centroids" may not be accurate when we set the minimum
      acceptable slice to be greater than 1. May need to change the name
      later.

    - ``radius``: the half-width of a square area of the pixels from which the
      centroids are calculated (pixel units).

      :Type: positive *float*
      :Required by: :func:`~HS_Centroids.find_centroids_using_template`

    - ``weight_factor``: An exponent to be applied to the pixel values of an
      image to be centroided.

      :Type: positive *float*
      :Required by: :func:`~HS_Centroids.find_centroids_using_template`

    - ``resolution``: The resolution of an image to be centroided.

      :Type: *list* of two elements
      :Updated by: :func:`~HS_Centroids.determine_resolution`

    - ``intensities``: The sums of the pixel values of the centroids.

      :Type: *ndarray*
      :Updated by: :func:`~HS_Centroids.find_centroids_using_template`

    - ``moments``: a list of x, y, x2 and y2 moments.

      :Type: *dict*
      :Updated by: :func:`~HS_Centroids.compute_moments`

    .. note:: The variables below are currently not in use.

    - ``half_offset``
    - ``corner_centroids``
    - ``centroids_x``
    - ``centroids_y``
    - ``spot_spacing``
    - ``spacing_tolerance_factor``
    - ``dev_tolerance``
    - ``is_template``
    - ``spots_are_valid``
    - ``expected_no_centroids``

    """

    def __init__(self):
        self.centroids = None
        self.corner_centroids = None
        self.centroids_x = None
        self.centroids_y = None
        self.spot_spacing = None
        self.slices_size = None
        self.spacing_tolerance_factor = 0.1
        self.dev_tolerance = 2
        self.threshold = None
        self.ot_modifier = 1.0
        self.otsu_threshold = None
        self.hsimage = None
        self.is_template = None
        self.spots_are_valid = None
        self.no_of_centroids = None
        self.expected_no_centroids = None
        self.rejected_centroids = None
        self.radius = 10
        self.average = None
        self.half_offset = None
        self.weight_factor = 2
        self.intensities = None
        self.max_pvs = None
        self.resolution = None
        self.moments = {}
        self.moments["x"] = None
        self.moments["y"] = None
        self.moments["x2"] = None
        self.moments["y2"] = None

    def average_centroids(self):
        """Average the centroids.

        :Requires: ``centroids``
        :Updates: ``average``

        """
        self.average = array([self.centroids[:, 0].mean(), self.centroids[:, 1].mean()])

    def find_centroids(self, refcents=None):
        """Calculate the centroids using the reference as the reference.

        If an array of the centroids ``refcents`` is given as an input parameter,
        it is used as a set of initial coordinates to calculate the centroids.

        If no parameters are given, the method will attempt to detect the spots
        in an image, then calculate the centroids.

        :uses: :func:`~HS_Centroids.find_centroids_from_image`,
               :func:`~HS_Centroids.find_centroids_using_template`
        :param refcents: N by 2 ndarray of the reference centroids
        :type refcents: *ndarray* or *None*

        """
        if refcents is None:
            self.find_centroids_from_image()
        else:
            self.find_centroids_using_template(refcents)

    def determine_resolution(self):
        """Determine the resolution of an image to be centroided.

        First checks if the instance variable ``hsimage`` is an instance
        of :class:`HS_Image <HWS.HS_Image.HS_Image>`, then obtains the
        shape of the array ``hsimage.original_image``.

        :Requires: ``hsimage.original_image``
        :Updates: ``resolution``

        """
        if self.hsimage.__class__.__name__ == "HS_Image":
            self.resolution = zeros(2)
            res = shape(self.hsimage.original_image)
            self.resolution[0] = res[1]
            self.resolution[1] = res[0]
        else:
            raise Exception("hsimage is not an instance of HS_Image.")

    def find_centroids_from_image(self, to_filter_spots=False, size_threshold=1):
        """Find the centroids of an image.

        The method determines the image thresholding pixel value by calling
        ``skf.threshold_otsu`` method, uses that value to
        threshold the image, then detects the spots in the image using the
        ``scipy`` methods ``ndimage.label`` and ``ndimage.find_objetcts``.

        The method then calls the instance method
        :func:`~HS_Centroids.find_centroids_using_template` to compute the
        centroids.

        Optionally, if ``to_filter_spots`` is set ``True``, the
        method rejects any detected objects whose slice size is equal or
        lower than ``size_threshold``.

        :param to_filter_spots: (optional) if ``True``, spots of small size are not in                                                 cluded as centroids.
        :type to_filter_spots: *boolean*
        :param size_threshold: (optional) the maximum spot slice size to be rejected by
                                          filtering.
        :type size_threshold:
        :Requires: ``hsimage.modified_image``
        :uses: :func:`~HS_Centroids.determine_resolution`
               :func:`~HS_Centroids.find_centroids_using_template`
        :Updates: ``otsu_threshold``, ``centroids``

        """
        self.determine_resolution()
        otsu_threshold = threshold_otsu(self.hsimage.modified_image)

        bi = self.hsimage.modified_image > otsu_threshold * self.ot_modifier
        ilabel, noc = scipy.ndimage.label(bi)

        spot_slices = scipy.ndimage.find_objects(ilabel)
        rejected_centroids = None
        # print noc, shape(spot_slices)

        # for debugging
        # self.spot_slices = spot_slices
        if to_filter_spots is True:
            peaks, rejected_centroids = self.calculate_slice_centres_filtered(
                spot_slices, size_threshold
            )
        else:
            peaks = self.calculate_slice_centres(spot_slices)

        # print peaks
        self.find_centroids_using_template(peaks)
        self.otsu_threshold = otsu_threshold
        if rejected_centroids is not None:
            self.rejected_centroids = rejected_centroids

    def calculate_slice_centres_filtered(self, spot_slices, size_threshold):
        """Determine the centres of the objects detected from an image and filter
        out those that are too small.

        The method is called optionally by :func:`~HS_Centroids.find_centroids_from_image`.

        :param spot_slices: slice objects returned by SciPy method ``find_objects``
        :type spot_slices: *list* of *tuples*
        :param size_threshold: (optional) the maximum spot slice size to be rejected by filtering
        :type size_threshold: non-negative *integer*
        :returns: coordinates of the detected objects and rejected objects
        :rtype: *tuple* of *ndarrays*

        """
        # noc = shape(spot_slices)[0]
        # peaks = zeros((noc,2))

        # counter = 0
        peaks = None
        rejected_centroids = None
        slices_size = None

        for sr in spot_slices:
            ss = (sr[0].stop - sr[0].start) * (sr[1].stop - sr[1].start)
            x = round((sr[1].start + sr[1].stop - 1) / 2.0)
            y = round((sr[0].start + sr[0].stop - 1) / 2.0)
            if ss > size_threshold:
                if peaks is None:
                    peaks = array([x, y])
                    slices_size = array([ss])
                else:
                    peaks = vstack((peaks, array([x, y])))
                    slices_size = vstack((slices_size, array([ss])))
            else:
                if rejected_centroids is None:
                    rejected_centroids = array([x, y])
                else:
                    # print rejected_centroids
                    rejected_centroids = vstack((rejected_centroids, array([x, y])))

        self.slices_size = slices_size.ravel()

        return peaks, rejected_centroids

    def calculate_slice_centres(self, spot_slices):
        """Determine the centres of the objects detected from an image.

        :param spot_slices: slice objects returned by SciPy method ``find_objects``
        :type spot_slices: *list* of *tuples*
        :returns: coordinates of the detected objects and rejected objects
        :rtype: *tuple* of *ndarrays*

        """
        # noc = shape(spot_slices)[0]
        # peaks = zeros((noc,2))

        # counter = 0
        peaks = None
        slices_size = None

        for sr in spot_slices:
            ss = (sr[0].stop - sr[0].start) * (sr[1].stop - sr[1].start)
            x = round((sr[1].start + sr[1].stop - 1) / 2.0)
            y = round((sr[0].start + sr[0].stop - 1) / 2.0)
            if peaks is None:
                peaks = array([x, y])
                slices_size = array([ss])
            else:
                peaks = vstack((peaks, array([x, y])))
                slices_size = vstack((slices_size, array([ss])))

        self.slices_size = slices_size.ravel()

        return peaks

    def find_centroids_using_template_old(self, refcents):
        """Calculate the centroids of an image using an array of the
        centroids as starting positions.

        :param refcents: an N by 2 array of the reference centroids (N = the number of
                         centroids)
        :type refcents: ndarray
        :Requires: ``hsimage.modified_image``, ``weight_factor``, ``radius``
        :uses: :func:`~HS_Centroids.determine_resolution`,
               :func:`~HS_Centroids.determine_range`,
               :func:`~HS_Centroids.calculate_centroid`,
        :Updates: ``centroids``, ``intensities``, ``max_pvs``,
                  ``no_of_centroids``
        """
        self.determine_resolution()
        noc = shape(refcents)[0]
        peaks = rint(refcents).astype(int)
        # im = self.hsimage.modified_image.copy()
        im = self.hsimage.modified_image.astype(float)
        if self.threshold != None:
            im[im < self.threshold] = 0

        wim = im ** self.weight_factor

        self.centroids = []
        self.intensities = []
        self.max_pvs = []

        for p in peaks:
            xr, yr = self.determine_range(p, self.radius)
            xg, yg = meshgrid(xr, yr)
            xg = xg.astype(int)
            yg = yg.astype(int)

            twim = wim[yg, xg]
            tim = im[yg, xg]
            ctr = self.calculate_centroid(xg, yg, twim)

            # print(ctr)
            # print(any(ctr == None))
            if any(ctr) == None:
                # self.centroids.append(ctr)
                newp = ctr
            else:
                newp = rint(ctr).astype(int)

            if (newp[0] == p[0]) and (newp[1] == p[1]):
                self.centroids.append(ctr)
            elif (newp[0] != None) and (newp[1] != None):
                xr, yr = self.determine_range(newp, self.radius)
                xg, yg = meshgrid(xr, yr)
                xg = xg.astype(int)
                yg = yg.astype(int)
                try:
                    twim = wim[yg, xg]
                except IndexError:
                    print(newp, xr, yr)
                tim = im[yg, xg]
                self.centroids.append(self.calculate_centroid(xg, yg, twim))

            # print p
            self.intensities.append(tim.sum())
            self.max_pvs.append(tim.max())

        self.centroids = array(self.centroids)
        self.intensities = array(self.intensities)
        self.max_pvs = array(self.max_pvs)
        self.no_of_centroids = noc

    def find_centroids_using_template(self, refcents):
        """Calculate the centroids of an image using an array of the
        centroids as starting positions.

        :param refcents: an N by 2 array of the reference centroids (N = the number of
                         centroids)
        :type refcents: ndarray
        :Requires: ``hsimage.modified_image``, ``weight_factor``, ``radius``
        :uses: :func:`~HS_Centroids.determine_resolution`,
               :func:`~HS_Centroids.determine_range`,
               :func:`~HS_Centroids.calculate_centroid`,
        :Updates: ``centroids``, ``intensities``, ``max_pvs``,
                  ``no_of_centroids``
        """
        self.determine_resolution()
        noc = shape(refcents)[0]
        peaks = rint(refcents).astype(int)
        # im = self.hsimage.modified_image.copy()
        im = self.hsimage.modified_image.astype(float)
        if self.threshold != None:
            im[im < self.threshold] = 0

        wim = im ** self.weight_factor

        self.centroids = []
        self.intensities = []
        self.max_pvs = []

        for p in peaks:
            xr, yr = self.determine_range(p, self.radius)
            xg, yg = meshgrid(xr, yr)
            xg = xg.astype(int)
            yg = yg.astype(int)

            twim = wim[yg, xg]
            tim = im[yg, xg]
            ctr = self.calculate_centroid(xg, yg, twim)

            # print(ctr)
            # print(any(ctr == None))
            if (ctr[0] == None) or (ctr[1] == None):
                self.centroids.append(ctr)
                newp = ctr
            else:
                newp = rint(ctr).astype(int)

            if (newp[0] == p[0]) and (newp[1] == p[1]):
                self.centroids.append(ctr)
            elif (newp[0] != None) and (newp[1] != None):
                xr, yr = self.determine_range(newp, self.radius)
                xg, yg = meshgrid(xr, yr)
                xg = xg.astype(int)
                yg = yg.astype(int)
                try:
                    twim = wim[yg, xg]
                except IndexError:
                    print(newp, xr, yr)
                tim = im[yg, xg]
                self.centroids.append(self.calculate_centroid(xg, yg, twim))

            # print p
            self.intensities.append(tim.sum())
            self.max_pvs.append(tim.max())

        self.centroids = array(self.centroids)
        self.intensities = array(self.intensities)
        self.max_pvs = array(self.max_pvs)
        self.no_of_centroids = noc

    def determine_range(self, p, r, resolution=None):
        """Determine the x and y ranges centred at p with length ``r``.

        It ensures that the range does not go over the maximum and
        minimum possible coordinate values (in pixel units).

        :param p: the x and y coordinate of the centre of the range to
                  be found.
        :type p: *list* or *ndarray* of two elements
        :param r: the distance from ``p``
        :type p: positive *float*
        :returns: two *ndarrays* for the ranges in x and in y directions
        :rtype: *tuple*

        """
        if resolution is None:
            resolution = self.resolution

        xmax = resolution[0]
        ymax = resolution[1]
        xr = arange(max(0, p[0] - r), min(p[0] + r + 1, xmax))
        yr = arange(max(0, p[1] - r), min(p[1] + r + 1, ymax))

        return xr, yr

    def calculate_centroid(self, xc, yc, tim):
        """Calculate the centroid of a region of an image.

        The method simply calcualtes the centre of mass of an image (or
        a region thereof) ``tim``, treating its pixel values like
        masses.

        :param xc: a 2D array of x coordinates, as generated by the numpy
                   method ``meshgrid``
        :type xc: *ndarray*
        :param yc: a 2D array of y coordinates, as generated by the numpy
                   method ``meshgrid``
        :type yc: *ndarray*
        :param tim: a 2D array of pixel values, whose shape must be the same as
                   ``xc`` and ``yc``
        :type tim: *ndarray*
        :returns: the calculated centroid (x and y coordinates)
        :rtype: *ndarray*

        """

        total_intensity = float(tim.sum())
        xcf = xc.astype(float)
        ycf = yc.astype(float)

        if total_intensity == 0:
            xct = None
            yct = None
        else:
            xct = (tim * xcf).sum() / total_intensity
            yct = (tim * ycf).sum() / total_intensity

        return array([xct, yct])

    def compute_moments(self):
        """Calculate the moments x, y, x2 and y2.

        :Requires: ``centroids``, ``intensities``
        :Updates: ``moments``

        """
        if (self.centroids == None) or (self.intensities == None):
            raise Exception(
                "Instance variables centroids and intensities "
                + "must be present for the moments calculation."
            )
        totI = self.intensities.sum()
        avC = dot(self.centroids.T, self.intensities) / totI
        self.moments["x"] = avC[0]
        self.moments["y"] = avC[1]
        self.moments["x2"] = (
            (self.centroids[:, 0] - self.moments["x"]) ** 2 * self.intensities
        ).sum() / totI
        self.moments["y2"] = (
            (self.centroids[:, 1] - self.moments["y"]) ** 2 * self.intensities
        ).sum() / totI

    def estimate_beam_width(self):
        """Calculate the estimate of the beam widths in x and y
        directions from the moment x2 and y2.

        :returns: the beam widths in x and y directions
        :rtype: *tuple*
        """
        self.calculate_moments()
        width_x = 4 * (self.second_moments["x2"]) ** 0.5
        width_y = 4 * (self.second_moments["y2"]) ** 0.5

        return (width_x, width_y)



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from copy import deepcopy
from numpy import *
from numpy.linalg import norm
from scipy import arctan
from HWS.HS_Centroids import *
from HWS.HS_Gradients import *
from HWS.HS_WFP import *


class HS_OFF2ON:
    """A Class to determine the angle of the input beam and the tilt of the
    camera to transform off-axis gradients to on-axis gradients, and
    calculates on-axis wavefront aberration.

    **Instance Variables**

    - ``magnification``: the magnification factor of the beam received by
      the camera. The default value is 1.

    - ``unitization_length``: the unitization length that is to be used for
      all properties that are instances of HS_WFP class.
      :func:`~HS_OFF2ON.enable_unitization` method sets the
      unitization_length of all such properties.

    - ``hsc_cref_off``: an instance of
      :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>` that holds the
      reference centroids obtained from a collimated (non-divergent) beam.

    - ``hsc_sref_off``: an instance of
      :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>` that holds the
      centroids from the reflection off a cold spherical mirror.

    - ``hsc_live_off``: an instance of
      :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>` that holds the
      centroids from the reflection off a heated spherical mirror.

    - ``hsg_sref_off``: an instance of
      :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>` that holds the
      gradients between ``cref`` and ``sref`` centroids.

    - ``hsg_live_off``: an instance of
      :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>` that holds the
      gradients between ``live`` and ``cref`` centroids.

    - ``hsg_live_sref_off``: an instance of
      :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>` that holds the
      gradients between ``live`` and ``sref`` centroids.

    - ``incidence_angle``: the angle of incidence between the input and the
      reflected beam. When ``sref`` centroids are obtained from a cold
      spherical mirror, this value can be calculated using the method
      :func:`~HS_OFF2ON.calculate_incidence_angle`. This value must either
      be calculated from off-axis data or set by a user before calculating
      on-axis gradients.

    - ``tilt_angle``: the angle of camera tilt. When ``sref`` centroids are
      obtained from a cold spherical mirror, this value can be calculated
      using the method :func:`~HS_OFF2ON.calculate_incidence_angle`. This value
      must be either calculated from off-axis data or set by a user before
      calculating on-axis gradients.

    - ``normal_vectors_sref_off``: an array of vectors normal to the
      surface of the cold mirror (POV off-axis camera).

    - ``normal_vectors_sref_on``: an array of vectors normal to the
      surface of the cold mirror (POV on-axis).

    - ``normal_vectors_live_off``: an array of vectors normal to the
      surface of the hot mirror (POV off-axis camera).

    - ``normal_vectors_live_on``: an array of vectors normal to the
      surface of the hot mirror (POV on-axis).

    - ``hswfp_sref_off``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the off-axis
      gradients between ``sref`` and ``cref``.

    - ``hswfp_live_off``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the off-axis
      gradients between ``live`` and ``cref``.

    - ``hswfp_live_sref_off``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the off-axis
      gradients between ``live`` and ``sref``.

    - ``hswfp_sref_on``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the on-axis
      gradients between ``sref`` and ``cref``.

    - ``hswfp_live_on``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the on-axis
      gradients between ``live`` and ``cref``.

    - ``hswfP_live_sref_on``: an instance of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`
      used to calculate the polynomial and Seidel coefficients from the on-axis
      gradients between ``live`` and ``sref``.

    """

    def __init__(self, hsc_cref_off=None, hsc_sref_off=None, hsc_live_off=None):
        self.magnification = 1.0
        self.unitization_length = None
        self.incidence_angle = None
        self.tilt_angle = None
        self.hsc_cref_off = hsc_cref_off
        self.hsc_sref_off = hsc_sref_off
        self.hsc_live_off = hsc_live_off
        self.hsg_sref_off = None
        self.hsg_live_off = None
        self.hsg_live_sref_off = None
        self.hsc_cref_on = None
        self.hsc_sref_on = None
        self.hsc_live_on = None
        self.hswfp_sref_off = HS_WFP()
        self.hswfp_live_off = HS_WFP()
        self.hswfp_live_sref_off = HS_WFP()
        self.hswfp_sref_on = HS_WFP()
        self.hswfp_live_on = HS_WFP()
        self.hswfp_live_sref_on = HS_WFP()
        self.normal_vectors_sref_off = None
        self.normal_vectors_sref_on = None
        self.normal_vectors_live_off = None
        self.normal_vectors_live_on = None

    def enable_unitization(self):
        """Enable the option to unitize the length scales for the WF coefficient
        calculations.

        This method turns on the option to use unitization for all
        properties that are instances of :class:`HS_WFP <HWS.HS_WFP.HS_WFP>`,
        and sets their ``unitization_length`` instance variables.

        A user can control unitization of each instances separately as
        well. This method is just for convenience when a user wants to
        enable unitization with the same unitization length all at once.

        :Requires: ``unitization_length``
        :Updates: ``to_unitize`` and ``unitization_length`` of ``hswfp_sref_off``,
                  ``hswfp_live_off``, ``hswfp_live_sref_off``, ``hswfp_sref_on``,
                  ``hswfp_live_on``, and ``hswfp_sref_live_on``

        """
        self.hswfp_sref_off.to_unitize = True
        self.hswfp_live_off.to_unitize = True
        self.hswfp_live_sref_off.to_unitize = True
        self.hswfp_live_sref_on.to_unitize = True
        self.hswfp_sref_on.to_unitize = True
        self.hswfp_live_on.to_unitize = True

        self.hswfp_sref_off.unitization_length = self.unitization_length
        self.hswfp_live_off.unitization_length = self.unitization_length
        self.hswfp_live_sref_off.unitization_length = self.unitization_length
        self.hswfp_sref_on.unitization_length = self.unitization_length
        self.hswfp_live_on.unitization_length = self.unitization_length
        self.hswfp_live_sref_on.unitization_length = self.unitization_length

    def compute_off(self):
        """Compute off-axis gradients and the polynomial coefficients.

        :Uses: :func:`~HS_WFP.compute_sref_off`,
               :func:`~HS_WFP.compute_live_off`

        """
        self.compute_sref_off()
        self.compute_live_off()

    def compute_incidence_angle(self):
        """Compute the angle of incidence of the beam upon a spherical mirror.

        Uses the seidel coefficients obtained from ``sref`` data to compute
        the angle of incidence of the input rays as they hit the mirror. ``seidel_coeffs``
        of ``hswfp_sref_off`` must be present before calling this method.

        :Requires: ``seidel_coeffs`` of ``hswfp_sref_off``
        :Updates: ``incidence_angle``

        """
        S = self.hswfp_sref_off.seidel_coeffs["spherical_power"]
        if S > 0:
            C = self.hswfp_sref_off.seidel_coeffs["cylindrical_power"]
        else:
            hswfp_n = deepcopy(self.hswfp_sref_off)
            hswfp_n.gradients = -hswfp_n.gradients
            hswfp_n.compute_poly_coeffs()
            hswfp_n.compute_seidel_coeffs()
            S = hswfp_n.seidel_coeffs["spherical_power"]
            C = hswfp_n.seidel_coeffs["cylindrical_power"]

        self.incidence_angle = arctan(sqrt(C / abs(S)))

    def compute_incidence_angle_old(self):
        """Compute the angle of incidence of the beam upon a spherical mirror.

        Uses the seidel coefficients obtained from ``sref`` data to compute
        the angle of incidence of the input rays as they hit the mirror. ``seidel_coeffs``
        of ``hswfp_sref_off`` must be present before calling this method.

        :Requires: ``seidel_coeffs`` of ``hswfp_sref_off``
        :Updates: ``incidence_angle``

        """
        S = self.hswfp_sref_off.seidel_coeffs["spherical_power"]
        C = self.hswfp_sref_off.seidel_coeffs["cylindrical_power"]

        self.incidence_angle = arctan(sqrt(C / abs(S)))

    def compute_tilt_angle(self):
        """Compute the tilt angle of the Hartmann sensor.

        Uses the seidel coefficients obtained from ``sref`` data to compute
        the tilt angle of the Harmann sensor.

        :Requires: ``seidel_coeffs`` of ``hswfp_sref_off``
        :Updates: ``tilt_angle``

        """
        if self.hswfp_sref_off.seidel_coeffs["spherical_power"] > 0:
            phi = self.hswfp_sref_off.seidel_coeffs["phi"]
        else:
            hswfp_n = deepcopy(self.hswfp_sref_off)
            hswfp_n.gradients = -hswfp_n.gradients
            hswfp_n.compute_poly_coeffs()
            hswfp_n.compute_seidel_coeffs()
            phi = self.hswfp_sref_off.seidel_coeffs["phi"]

        if phi > 0:
            self.tilt_angle = pi / 2 - phi
        else:
            self.tilt_angle = -pi / 2 - phi

    def compute_on(self):
        """Compute the on-axis gradients and the wavefront coefficients.

        This is the wrapper method for the computation of the on-axis
        gradients and the wavefront coefficients.

        """
        self.compute_normal_vectors()
        self.compute_gradients_on()
        self.compute_centroids_on()
        self.compute_coeffs_on()

    def compute_normal_vectors(self):
        """Compute sref and live normal bectors."""
        self.compute_normal_vectors_sref()
        self.compute_normal_vectors_live()

    def compute_gradients_on(self):
        """Calculate on-axis gradients for all ``hswfp`` instance variables.

        :Uses: :func:`~HS_OFF2ON.compute_bradients_sref_on`,
               :func:`~HS_OFF2ON.compute_bradients_live_on`,
               :func:`~HS_OFF2ON.compute_bradients_live_sref_on`
        """
        self.compute_gradients_sref_on()
        self.compute_gradients_live_on()
        self.compute_gradients_live_sref_on()

    def compute_centroids_on(self):
        """Transforms off-axis centroids to on-axis centroids.

        :Requires: ``incidence_angle``, ``tilt_angle``, ``hsg_sref_off``,
                   ``hsg_live_off``
        :Updates: ``centroids`` of ``hsc_cref_on``, ``hswfp_live_sref_on``,
                  ``hswfp_live_sref_on``

        """
        self.hsc_cref_on = HS_Centroids()
        self.hsc_sref_on = HS_Centroids()
        self.hsc_live_on = HS_Centroids()
        origin = self.hsg_sref_off.origin
        lever_arm = self.hsg_sref_off.lever_arm
        pixel_size = self.hsg_sref_off.pixel_size
        ia = self.incidence_angle
        ta = self.tilt_angle
        rotmat_cam = matrix([[cos(ta), -sin(ta)], [sin(ta), cos(ta)]])

        # cref_on is constructed by transforming cref_off centroids
        cc = transpose(matrix(self.hsg_sref_off.gradients[:, [2, 3]]))
        cc = array(transpose(rotmat_cam * cc))
        cc[:, 0] = cc[:, 0] / cos(ia)

        # sc = self.hsg_sref_off.gradients[:,[0,1]] + self.hsg_sref_off.gradients[:,[2,3]]
        # sc = rotmat_cam*transpose(matrix(sc))
        # sc = array(transpose(rotmat_cam*sc))

        # lc = self.hsg_live_off.gradients[:,[0,1]] + self.hsg_live_off.gradients[:,[2,3]]
        # lc = rotmat_cam*transpose(matrix(lc))
        # lc = array(transpose(rotmat_cam*lc))

        # sref_on and live_on centroids are constructed from the on-axis gradients
        sc = self.hswfp_sref_on.gradients * lever_arm + cc
        lc = self.hswfp_live_on.gradients * lever_arm + cc

        self.hswfp_sref_on.centroids = cc
        self.hswfp_live_on.centroids = cc
        self.hswfp_live_sref_on.centroids = sc
        # self.hswfp_live_sref_on.centroids = self.hsc_sref_on.centroids

        self.hsc_cref_on.centroids = cc / pixel_size + origin
        self.hsc_sref_on.centroids = sc / pixel_size + origin
        self.hsc_live_on.centroids = lc / pixel_size + origin

    def compute_coeffs_on(self):
        """Compute the on-axis polynomial coefficients for all ``hswfp`` instance
        variables.

        :Requires: ``hswfp_sref_on``, ``hswfp_live_on``, ``hswfp_live_sref_on``

        """
        self.hswfp_sref_on.compute_poly_coeffs()
        self.hswfp_sref_on.compute_seidel_coeffs()

        self.hswfp_live_on.compute_poly_coeffs()
        self.hswfp_live_on.compute_seidel_coeffs()

        self.hswfp_live_sref_on.compute_poly_coeffs()
        self.hswfp_live_sref_on.compute_seidel_coeffs()

    def compute_sref_off(self):
        """Compute the off-axis gradients and the polynomial coefficients from
        ``sref`` data.

        This method calcualtes the gradients between the reference
        (``cref``) and off-axis ``sref`` centroids (the centroids that are
        supposed to be due to the rays refelcted by a cold spherical
        mirror), then calculates the corresponding polynomial coefficients.

        :Requires: ``hsc_cref_off``, ``hsc_sref_off``
        :Updates: ``hsg_sref_off``, ``hswfp_sref_off``

        """
        self.hsg_sref_off = HS_Gradients(self.hsc_cref_off, self.hsc_sref_off)

        self.hsg_sref_off.magnification = self.magnification

        self.hsg_sref_off.construct_gradients()

        self.hswfp_sref_off.centroids = self.hsg_sref_off.gradients[:, [2, 3]]
        self.hswfp_sref_off.gradients = self.hsg_sref_off.gradients[:, [0, 1]]

        self.hswfp_sref_off.compute_poly_coeffs()
        self.hswfp_sref_off.compute_seidel_coeffs()

    def compute_live_off(self):
        """Compute the off-axis gradients and the polynomial coefficients from
        ``live`` data.

        This method calcualtes the gradients between the reference
        (``cref``) and off-axis ``live`` centroids (the centroids that are
        supposed to be due to the rays refelcted by a thermally agitated
        spherical mirror), then calculates the corresponding polynomial
        coefficients.

        :Requires: ``hsc_cref_off``, ``hsc_live_off``
        :Updates: ``hsg_live_off``, ``hswfp_live_off``

        """
        self.hsg_live_off = HS_Gradients(self.hsc_cref_off, self.hsc_live_off)
        self.hsg_live_sref_off = HS_Gradients(self.hsc_sref_off, self.hsc_live_off)

        self.hsg_live_off.magnification = self.magnification
        self.hsg_live_sref_off.magnification = self.magnification

        self.hsg_live_off.construct_gradients()
        self.hsg_live_sref_off.construct_gradients()

        self.hswfp_live_off.centroids = self.hsg_live_off.gradients[:, [2, 3]]
        self.hswfp_live_off.gradients = self.hsg_live_off.gradients[:, [0, 1]]

        self.hswfp_live_sref_off.centroids = self.hsg_live_sref_off.gradients[:, [2, 3]]
        self.hswfp_live_sref_off.gradients = self.hsg_live_sref_off.gradients[:, [0, 1]]

        self.hswfp_live_off.compute_poly_coeffs()
        self.hswfp_live_off.compute_seidel_coeffs()

        self.hswfp_live_sref_off.compute_poly_coeffs()
        self.hswfp_live_sref_off.compute_seidel_coeffs()

    def compute_normal_vectors_sref(self):
        """Compute the vecotrs normal to the reflecting surface of a cold spherical mirror.

        This method computes the vectors normalised and perpendicular to the
        reflecting surface from the off-axis ``sref`` data, then transforms
        them to the on-axis normal vectors.

        :Requires: ``tilt_angle``, ``incidence_angle``, ``hswfp_sref_off``
        :Updates: ``normal_vectors_sref_off``, ``normal_vectors_sref_on``

        """
        ta = self.tilt_angle
        ia = self.incidence_angle
        centroids = self.hswfp_sref_off.centroids
        gradients = self.hswfp_sref_off.gradients
        noc = shape(centroids)[0]
        # Rotate the centroids and gradients to align camera y axis
        # to master y axis
        rotmat_cam = matrix([[cos(ta), -sin(ta)], [sin(ta), cos(ta)]])
        grads_r = transpose(rotmat_cam * transpose(matrix(gradients)))

        # Make vectors in the direction of input rays from POV of camera
        in_angle = -ia * 2
        rotmat_in = matrix(
            [
                [cos(in_angle), 0, sin(in_angle)],
                [0, 1, 0],
                [-sin(in_angle), 0, cos(in_angle)],
            ]
        )
        z_neg = matrix([[0], [0], [-1]])
        z_neg_rot = transpose(rotmat_in * z_neg)

        n_i_POVcam = kron(ones((noc, 1)), array(z_neg_rot))

        # Determine the vector normal to the reflecting surface
        r, c = shape(grads_r)
        n_l_unscaled = ones((r, c + 1))
        n_l_unscaled[:, :-1] = grads_r

        n_l_norm = norm(n_l_unscaled, axis=1)
        n_l_POVcam = n_l_unscaled / n_l_norm[:, None]

        n_s_unscaled = n_l_POVcam - n_i_POVcam
        n_s_norm = norm(n_s_unscaled, axis=1)
        n_s_POVcam = n_s_unscaled / n_s_norm[:, None]

        self.normal_vectors_sref_off = n_s_POVcam

        # self.compute_incidence_angle()
        # ia = self.incidence_angle
        self.normal_vectors_sref_on = zeros((noc, 3))
        rotmat_m = matrix([[cos(ia), 0, sin(ia)], [0, 1, 0,], [-sin(ia), 0, cos(ia)],])
        for k in range(noc):
            nv_k = matrix(n_s_POVcam[k, :])
            res_mat = transpose(rotmat_m * transpose(nv_k))
            self.normal_vectors_sref_on[k, :] = array(res_mat)[0, :]

    def compute_normal_vectors_live(self):
        """Compute the vecotrs normal to the reflecting surface of a thermally agitated
        spherical mirror.

        This method computes the vectors normalised and perpendicular to the
        reflecting surface from the off-axis ``live`` data,
        then transforms them to the on-axis normal vectors.

        :Requires: ``tilt_angle``, ``incidence_angle``, ``hswfp_live_off``
        :Updates: ``normal_vectors_live_off``, ``normal_vectors_live_on``
        """
        ta = self.tilt_angle
        ia = self.incidence_angle
        centroids = self.hswfp_live_off.centroids
        gradients = self.hswfp_live_off.gradients
        noc = shape(centroids)[0]
        # Rotate the centroids and gradients to align camera y axis
        # to master y axis
        rotmat_cam = matrix([[cos(ta), -sin(ta)], [sin(ta), cos(ta)]])
        grads_r = transpose(rotmat_cam * transpose(matrix(gradients)))
        cents_r = transpose(rotmat_cam * transpose(matrix(centroids)))
        # Make vectors in the direction of input rays from POV of camera
        in_angle = -ia * 2
        rotmat_in = matrix(
            [
                [cos(in_angle), 0, sin(in_angle)],
                [0, 1, 0],
                [-sin(in_angle), 0, cos(in_angle)],
            ]
        )
        z_neg = matrix([[0], [0], [-1]])
        z_neg_rot = transpose(rotmat_in * z_neg)
        n_i_POVcam = kron(ones((noc, 1)), z_neg_rot)

        # Determine the vector normal to the reflecting surface
        r, c = shape(grads_r)
        n_l_unscaled = ones((r, c + 1))
        n_l_unscaled[:, :-1] = grads_r

        n_l_norm = norm(n_l_unscaled, axis=1)
        n_l_POVcam = n_l_unscaled / n_l_norm[:, None]

        n_s_unscaled = n_l_POVcam - n_i_POVcam
        n_s_norm = norm(n_s_unscaled, axis=1)
        n_s_POVcam = n_s_unscaled / n_s_norm[:, None]

        self.normal_vectors_live_off = n_s_POVcam

        # self.compute_incidence_angle()
        # ia = self.incidence_angle
        self.normal_vectors_live_on = zeros((noc, 3))
        rotmat_m = matrix([[cos(ia), 0, sin(ia)], [0, 1, 0,], [-sin(ia), 0, cos(ia)],])
        for k in range(noc):
            nv_k = matrix(n_s_POVcam[k, :])
            res_mat = transpose(rotmat_m * transpose(nv_k))
            self.normal_vectors_live_on[k, :] = array(res_mat)[0, :]

    def compute_gradients_sref_on(self):
        """Compute the on-axis gradients due to the reflection from the cold
        spherical mirror.

        This method constructs the gradients one would obtain on-axis, from the array
        ``normal_vectors_sref_on``.

        :Requires: ``normal_vectors_sref_on``
        :Updates: ``hswfp_sref_on``

        """
        n_i = array([0, 0, -1])
        noc = shape(self.normal_vectors_sref_on)[0]
        n_l = zeros((noc, 3))
        n_s = self.normal_vectors_sref_on
        # reflect n_i w.r.t. n_s to get n_l
        for k in range(noc):
            n_l[k, :] = n_i - 2 * (sum(n_i * n_s[k, :])) * n_s[k, :]

        gr = zeros((noc, 2))
        gr[:, 0] = n_l[:, 0] / n_l[:, 2]
        gr[:, 1] = n_l[:, 1] / n_l[:, 2]

        self.hswfp_sref_on.gradients = gr

    def compute_gradients_live_on(self):
        """Compute the on-axis gradients due to the reflection from the thermally
        agitated spherical mirror.

        This method constructs the gradients one would obtain on-axis, from the array
        ``normal_vectors_live_on``.

        :Requires: ``normal_vectors_live_on``
        :Updates: ``hswfp_live_on``

        """
        n_i = array([0, 0, -1])
        noc = shape(self.normal_vectors_live_on)[0]
        n_l = zeros((noc, 3))
        n_s = self.normal_vectors_live_on
        # reflect n_i w.r.t. n_s to get n_l
        for k in range(noc):
            n_l[k, :] = n_i - 2 * (sum(n_i * n_s[k, :])) * n_s[k, :]

        gr = zeros((noc, 2))
        gr[:, 0] = n_l[:, 0] / n_l[:, 2]
        gr[:, 1] = n_l[:, 1] / n_l[:, 2]

        self.hswfp_live_on.gradients = gr

    def compute_gradients_live_sref_on(self):
        """Compute on-axis gradients between live and sref.

        :Requires: ``gradients`` of ``hswfp_live_sref_on``, ``hswfp_live_on``, and
                   ``hswfp_sref_on``
        :Updates: ``hswfp_live_sref_on``

        """
        self.hswfp_live_sref_on.gradients = (
            self.hswfp_live_on.gradients - self.hswfp_sref_on.gradients
        )


# Won Kim 7/11/2018 3:49:06 PM



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from epics import PV
from numbers import Number
from numpy import *


class HS_PV(PV):
    """A class to interact with an EPICS process variable.

    ``HS_PV`` inherits the class ``PV`` of ``epics`` package (also called
    PyEpics), and implements type checking of the values before they are entered
    into an EPICS channel.

    All the methods of the class ``PV`` are hence available to ``HS_PV``, and a
    user is encouraged to consult PyEpics documentation for the full description
    of the package.

    Currently :func:`~HS_PV.check_type()` is the only method added to this class.
    The method :func:`~HS_PV.put()` is extended to incorporate the type checking
    done by :func:`~HS_PV.check_type()`.

    **Instantiation**

    The simplest way to instantiate the class is just provide ``record_name`` of
    a process variable. The created instance then acts just like ``PV`` class of
    ``pyepics``.

    If ``data_type`` is also given when instantiating, the instance method
    :func:`~HS_PV.put` checks if the value to be put is of the correct type using
    :func:`~HS_PV.check_type`.

    If the value is valid, then ``put`` method of ``PV`` class is called. If not,
    the put request is rejected with a warning message, and the value of the
    process variable stays as it was.

    The accepted names for the data types that can be checked and the meaning of
    each data type are implemented as *dict* ``data_type_verbose``:

    - ``boolean``: ``True`` or ``False``
    - ``pint``: positive *integer*
    - ``nnint``: non-negative *integer*
    - ``int``: *integer*
    - ``string``: *string*
    - ``pfloat``: positive real number
    - ``nnfloat``: non-negative real number
    - ``float``: any real number

    **Instance Variables**

    As this class inherits ``PV`` of ``pyepics``, there are many more instance
    variables that ``HS_PV``

    - ``data_type``: the data type of the process variable

      In order to check the data type, it must be one of the following:
      'boolean', 'pint', 'nnint', 'int', 'string', 'pfloat', 'nnfloat', or
      'float'.

      Setting ``data_type`` to something else won't raise an error but won't
      trigger data type checking.

      :Type: *string*
      :Required by: :func:`~HS_PV.put`, :func:`~HS_PV.check_type`

    - ``data_type_verbose``: *dict* of verbose descriptions of ``data_type`` values.

      This variable is used to produce an appropriate warning message when type
      checking yields ``False``.

      :Type: *dict*
      :Required by: :func:`~HS_PV.put`

    - ``type_checking``: *boolean* to indicate whether to check the data type or not

      :Type: *boolean*
      :Required by: :func:`~HS_PV.put`


    """

    def __init__(self, record_name, data_type=None):

        PV.__init__(self, record_name)

        self.default_value = None
        self.last_value = None
        self.data_type = data_type
        self.data_type_verbose = {}

        self.data_type_verbose["boolean"] = "boolean"
        self.data_type_verbose["pint"] = "positive integer"
        self.data_type_verbose["nnint"] = "non-negative integer"
        self.data_type_verbose["int"] = "integer"
        self.data_type_verbose["string"] = "string"
        self.data_type_verbose["pfloat"] = "positive real number"
        self.data_type_verbose["nnfloat"] = "non-negative real number"
        self.data_type_verbose["float"] = "real number"
        self.type_checking = False

        if (data_type is None) or (data_type == "waveform"):
            self.type_checking = False
        else:
            self.type_checking = True

    def put(
        self,
        value,
        wait=False,
        timeout=30.0,
        use_complete=False,
        callback=None,
        callback_data=None,
    ):
        """Check and enter a value into EPICS PV.

        See the PyEpics documentation for the ``pv.put`` method for the explanations
        of the input parameters other than ``value``.

        When the instance variable ``check_type`` is ``True``, this method calls


        :param value: the value of the process variable to be put into EPICS channel
        :Requires: ``type_checking``
        :Uses: (optional) :func:`~HS_PV.check_type`

        """

        if self.type_checking == True:
            if self.check_type(value):
                return PV.put(
                    self, value, wait, timeout, use_complete, callback, callback_data
                )
            else:
                print("**Warning** " + self.pvname)
                print(
                    "Not a valid data type: "
                    + "the record must be of the type "
                    + self.data_type_verbose[self.data_type]
                    + "."
                )
                print("The record value will not be changed.")
        else:
            return PV.put(
                self, value, wait, timeout, use_complete, callback, callback_data
            )
            self.last_value = value

    # def get(self,
    #         count=None,
    #         as_string=False,
    #         as_numpy=True,
    #         timeout=None,
    #         use_monitor=True):
    #     val = PV.get(self,count,as_string,as_numpy,timeout,use_monitor)

    #     if self.type_checking == True:
    #         if self.check_type(val):
    #             return val
    #         else:
    #             print ('**Warning** ' + self.pvname)
    #             print \
    #                 ('Not a valid data type: ' + \
    #                  'the record must be of the type ' + \
    #                  self.data_type_verbose[self.data_type] + '.')
    #             print('The record value will be reverted the last valid value.')

    def check_type(self, value):
        """Check if the value to be put is of correct data type

        :param value: the value to be checked
        :Returns: ``True`` or ``False``
        :rtype: *boolean*

        """

        if size(value) == 1:
            if self.data_type == "boolean":
                return (value == 0) or (value == 1)
            elif self.data_type == "pint":
                if isinstance(value, Number):
                    return (value == int(value)) and (value > 0)
                else:
                    return False
            elif self.data_type == "nnint":
                if isinstance(value, Number):
                    return (value == int(value)) and (value >= 0)
                else:
                    return False
            elif self.data_type == "int":
                return isinstance(value, Number) and (value == int(value))
            elif self.data_type == "pfloat":
                return isinstance(value, Number) and (value > 0)
            elif self.data_type == "nnfloat":
                return isinstance(value, Number) and (value >= 0)
            elif self.data_type == "float":
                return isinstance(value, Number) and not (isinstance(value, complex))
            elif self.data_type == "string":
                return isinstance(value, str)
        else:
            # TODO: revise to check waveform record as necessary. At the moment
            # the method returns False if the value is an array.

            return False



import HWS
import numpy as np
import pytest

from HWS.HS_Centroids import HS_Centroids
from HWS.HS_Image import HS_Image
import matplotlib.pyplot as plt


@pytest.mark.parametrize('N', [2, 4, 7])
@pytest.mark.parametrize('xc', [100, 300.333])
@pytest.mark.parametrize('yc', [100, 300.333])
@pytest.mark.parametrize('dx', [40, 33.33])
@pytest.mark.parametrize('dy', [40, 33.33])
def test_centroiding(N, xc, yc, dx, dy):
    x = np.arange(0, 900)
    y = np.arange(0, 1101)
    X, Y = np.meshgrid(x, y)

    image = np.zeros_like(X, dtype=float)
    to_test = []
    for n in range(1, N+1):
        dI = 100*np.exp(-((X-xc-dx*n)/10)**2 - ((Y-yc-dy*n)/10)**2)
        to_test.append([xc+dx*n, yc+dy*n])
        image += dI
        
    to_test = np.array(to_test, dtype=float)
    print("IN", to_test)
    hsi = HS_Image()
    hsi.original_image = image.astype(float)
    hsi.background = 0
    hsi.process_image()

    hsc = HS_Centroids()
    hsc.hsimage = hsi
    hsc.radius=20
    hsc.find_centroids_from_image()
    print("OUT", hsc.centroids)
    print("RELERR", abs(to_test - hsc.centroids)/abs(to_test))
    assert(len(hsc.centroids) == N)
    assert(np.all(abs(to_test - hsc.centroids)/abs(to_test) < 0.01))


"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.

.. note:: This module is a collection of methods that I found sometimes useful, 
          and is always "experimental". As such most methods do not yet have proper
          docstring description.

"""

import sys
from numpy import *
import matplotlib.pyplot as plt
from scipy.interpolate import griddata

if "cv2" in sys.modules:
    import cv2


def setup_grid(limits, gspacing):
    """Set up square mesh grids for wavefront construction.

    :param limits: the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :Returns: *ndarrays* of x and y grids
    :rtype: *tuple*

    """

    x1, x2 = limits["x"]
    y1, y2 = limits["y"]

    xr = arange(x1, x2 + gspacing, gspacing)
    yr = arange(y1, y2 + gspacing, gspacing)

    if xr[-1] > x2:
        xr = xr[:-1]

    if yr[-1] > y2:
        yr = yr[:-1]

    return meshgrid(xr, yr)


def save_image_frame(name, iframe):
    cv2.imwrite(name, iframe)


def read_image_frame(name):
    im = cv2.imread(name, cv2.IMREAD_ANYDEPTH)
    return im.astype(int)


def find_indices_near(hsc, c_ind, radius=None):
    """Find the centroids near a centroid.

    The method finds all centroids contained in ``hsc``, an instance of
    :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>``, that are within a
    distance from the centroid whose index is ``c_ind``.

    The distance is taken from ``radius``, and if it is not supplied, the
    method selects a value to be 1.25 times the smallest separation between
    the centroids labelled ``c_ind`` and its nearest centroid.

    Note that radius and the coordinates of the centroids in hsc must be in
    the same metric unit.

    The method returns the indices of the centroids thus found.

    """
    # Check if c_ind is valid relative to hsc
    noc = shape(hsc.centroids)[0]

    if c_ind not in arange(hsc.no_of_centroids):
        errstr = (
            "The centroid index c_ind is not consistent with "
            + "the number of centroids in hsc."
        )
        raise Exception(errstr)

    # Get the coordinates of the centre centroid.
    centre_c = hsc.centroids[c_ind]
    d = linalg.norm(hsc.centroids - centre_c, axis=1)

    # Determine the radius value to use if it is not supplied as an input.
    if radius is None:
        # Choose the second lowest value as the first lowest would be zero
        # (distance between the centre centroid and itself).
        radius = sort(d)[1] * 1.25
    elif radius < 0:
        raise Exception("radius must not be negative.")

    # Get the indices of the centroids that are within a radius from the
    # centre centroid.
    n_ind = where(d[:] < radius)

    return n_ind

    # Calculate distances between the centre centroid and the rest


def label_centroids_near_origin(hsc):
    anchors_index = argpartition(-hsc.slices_size, kth=2)[0:2]
    anchor_centroids = hsc.centroids[anchors_index]
    anchor_mean = anchor_centroids.mean(axis=0)
    origin_index = argmin(linalg.norm(hsc.centroids - anchor_mean, axis=1))
    origin_centroid = hsc.centroids[origin_index]
    nominal_sep = linalg.norm(anchor_centroids[0, :] - anchor_centroids[1, :]) / 2
    dist_fo = linalg.norm(hsc.centroids - origin_centroid, axis=1)
    cno_index = argpartition(dist_fo, kth=7)[0:7]
    cno_coord_fo = hsc.centroids[cno_index] - origin_centroid

    ri = {}
    ri["origin"] = origin_index
    anchor_from_origin = anchor_centroids - origin_centroid

    if abs(anchor_from_origin[0, 0]) > abs(anchor_from_origin[0, 1]):
        coord_to_label = 0
        anchor_types = ["left anchor", "right anchor"]
    else:
        coord_to_label = 1
        anchor_types = ["down anchor", "up anchor"]

    rc = hsc.centroids
    for ii in range(2):
        if anchor_from_origin[ii, coord_to_label] < 0:
            ri[anchor_types[0]] = anchors_index[ii]
        else:
            ri[anchor_types[1]] = anchors_index[ii]

    # Label the remaining 4 centroids near origin
    riv = ri.values()
    cc = 0
    for ii in cno_index:
        if ii not in riv:
            if cno_coord_fo[cc, 0] > 0:
                if cno_coord_fo[cc, 1] > 0:
                    ri["upper right"] = ii
                else:
                    ri["lower right"] = ii
            else:
                if cno_coord_fo[cc, 1] > 0:
                    ri["upper left"] = ii
                else:
                    ri["lower left"] = ii

        cc = cc + 1

    return ri


def interpolate_centroids_intensities(hsc):
    # WK: If the interpolated image look odd, it could be due to invalid
    # centroids that is more likely to occur if one lowers the otsu
    # threshold to get more centroids. It is recommended to check the
    # centroids and to filter centroids before using this method.

    x_max, x_min = (
        round((hsc.centroids[:, 0]).max()),
        round((hsc.centroids[:, 0]).min()),
    )

    y_max, y_min = (
        round((hsc.centroids[:, 1]).max()),
        round((hsc.centroids[:, 1]).min()),
    )

    limits = {}

    limits["x"] = array([x_min, x_max])
    limits["y"] = array((y_min, y_max))

    xg, yg = setup_grid(limits, 1)
    xc, yc = (xg[0, :], yg[:, 0])

    im_intp = griddata(hsc.centroids, hsc.intensities, (xg, yg), method="cubic")

    return im_intp


def read_and_average_image_frames(prefix, first=1, nof=1):
    pass


def generate_spot_pattern(
    centroids, size=(1024, 1024), pixel_size=12e-6, spot_radius=30e-6
):

    ss = spot_radius / pixel_size
    ssb = ss * 2
    image = zeros(size)

    for ctr in centroids:
        pos_is_valid = (
            ctr[0] > 0 + 20
            and ctr[0] < size[0] - 20
            and ctr[1] > 0 + 20
            and ctr[1] < size[1] - 20
        )

        if pos_is_valid:
            xmin = array([0, ctr[0] - ssb]).max().astype(int)
            xmax = array([size[0], ctr[0] + ssb + 1]).min().astype(int)
            xr = arange(xmin, xmax)

            ymin = array([0, ctr[1] - ssb]).max().astype(int)
            ymax = array([size[1], ctr[1] + ssb + 1]).min().astype(int)
            yr = arange(ymin, ymax)

            xg, yg = meshgrid(xr, yr)
            image[yg, xg] = exp(
                -(
                    (
                        ((xg - ctr[0]) / (2.0 * ss)) ** 2
                        + ((yg - ctr[1]) / (2.0 * ss)) ** 2
                    )
                    ** (10 / 2.0)
                )
            )

    return image


def generate_hex_grid_old():
    hole_spacing = 470e-6
    angle = 0
    n = 35
    (X, Y) = meshgrid(arange(n + 1), arange(n + 1))
    X = X * hole_spacing
    Y = Y * hole_spacing
    n = shape(X)[0]
    X = X * (3.0 ** 0.5 / 2)
    Y = Y + tile(array([0, 0.5 * hole_spacing]), [n, n / 2])
    X_s = (X / 12e-6) + 30
    Y_s = (Y / 12e-6) + 15
    X_r = zeros([n, n])
    Y_r = zeros([n, n])

    for ii in arange(n):
        for jj in arange(n):
            X_r[ii, jj] = cos(angle) * X_s[ii, jj] - sin(angle) * Y_s[ii, jj]
            Y_r[ii, jj] = sin(angle) * X_s[ii, jj] + cos(angle) * Y_s[ii, jj]

    pos = array([X_r.flatten(), Y_r.flatten()]).T
    ind = lexsort((pos[:, 1], pos[:, 0]))
    return pos[ind]


def generate_hex_grid(
    length=[1024, 1024], separation=37, orientation=0, padding=[30, 30]
):
    sp_s = separation
    sp_l = sqrt(3) * separation

    if orientation == 0:
        xspacing = sp_s
        yspacing = sp_l
    else:
        xspacing = sp_l
        yspacing = sp_s

    # outer x and y coordinates
    xr_o = arange(start=padding[0], stop=length[0] - padding[0], step=xspacing)
    yr_o = arange(start=padding[1], stop=length[1] - padding[1], step=yspacing)

    # inner x and y coordinates
    xr_i = arange(
        start=padding[0] + xspacing / 2.0,
        stop=length[0] - padding[0] - xspacing / 2.0,
        step=xspacing,
    )
    yr_i = arange(
        start=padding[1] + yspacing / 2.0,
        stop=length[1] - padding[1] - yspacing / 2.0,
        step=yspacing,
    )

    xg_o, yg_o = meshgrid(xr_o, yr_o)
    xg_i, yg_i = meshgrid(xr_i, yr_i)

    c_o = array(matrix([xg_o.ravel(), yg_o.ravel()]).T)
    c_i = array(matrix([xg_i.ravel(), yg_i.ravel()]).T)

    ctr = concatenate((c_o, c_i), axis=0)
    idx = lexsort((ctr[:, 1], ctr[:, 0]))
    return ctr[idx]


def generate_and_plot_centroids():
    ctr = generate_hex_grid()
    plt.scatter(ctr[:, 0], ctr[:, 1], marker="+")
    plt.gca().set_aspect("equal")
    plt.xlim(0, 1024)
    plt.ylim(0, 1024)

    for n in arange(shape(ctr)[0]):
        plt.text(ctr[n, 0] + 5, ctr[n, 1] + 5, str(n + 1), fontsize=8)

    plt.draw()

    return ctr



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

import glob
import os
import sys
from numpy import *
from numbers import Number

# from scipy import *

__blur_cache = {}

class HS_Image(object):
    """A class to interact with the images taken from a camera.

    ``HS_Image`` is a class that is used to read the image frames and to process
    them prior to centroiding.

    **Instantiation**

    The class has two main options to read the image frames, specified by the
    instance variable ``location_type``.

    If ``location_type`` is set to 'file', then ``no_of_frames`` is set to 1, and
    the instance is supposed to interact with a single file. The instance variable
    ``location`` should then be the corresponding full-path file name.

    If ``location_type`` is set to 'folder', more than one image frame file can be
    read and averaged to produce an image. Basic preparation for this process is
    performed by the method :func:`~HS_Image.set_acquisition_parameters`.

    An example way to use the class in 'folder' mode is, with ``location`` properly
    set to the folder in which the image frame files are stored, to set ``fprefix``
    (e.g., 'image') to be the file prefix of the files to be read, and set
    ``no_of_frames`` to a positive *integer* value (e.g., 10).

    Alternatively one can also set ``first_frame`` and ``last_frame`` as positive
    *integers*, ensuring that ``first_frame`` is smaller than ``last_frame``. The
    instance will then read and average that many files (from ``first_frame`` to
    ``last_frame``).

    **Instance Variables**

    - ``location_type``: the type of the ``location``, either 'file' or 'folder'

      :Type: *string*
      :Required by: :func:`~HS_Image.read_image`

    - ``location``: the path to a folder with image files to be read or the full path
      name of an image file to be read.

      :Type: *string*
      :Required by: :func:`~HS_Image.read_image`

    - ``first_frame``: a positive *integer* specifying the index of the first image
      frame file to be read

      This variable is applicable only when ``location_type`` is 'folder'.

      When ``fprefix`` is also given, ``first_frame`` is the index of the first file
      to be read with that file prefix, from a folder specified by ``location``.

      When ``fprefix`` is not given ``first_frame`` is the index of the first file to
      be read from a folder specified by ``location``.

      :Type: positive *integer*
      :Required by: (optional) :func:`~HS_Image.set_acquisition_parameters`,
                    :func:`~HS_Image.read_and_average_frames`
      :Updated by: (optional) :func:`~HS_Image.set_acquisition_parameters`,
                   :func:`~HS_Image.read_single_frame`

    - ``last_frame``: a positive *integer* specifying the index of the last image
      frame file to be read

      This variable is applicable only when ``location_type`` is 'folder'.

      When ``fprefix`` is also given, ``last_frame`` is the index of the last file
      to be read with that file prefix, from a folder specified by ``location``.

      When ``fprefix`` is not given ``last_frame`` is the index of the last file to
      be read from a folder specified by ``location``.

      :Type: positive *integer*
      :Required by: (optional) :func:`~HS_Image.set_acquisition_parameters`,
                    :func:`~HS_Image.read_and_average_frames`
      :Updated by: (optional) :func:`~HS_Image.set_acquisition_parameters`,
                   :func:`~HS_Image.read_single_frame`

    - ``no_of_frames``: the number of the image frame files to be read.

      :Type: positive *integer*
      :Required by: :func:`~HS_Image.set_acquisition_parameters` (optional),
                    :func:`~HS_Image.read_and_average_frames`
      :Updated by: :func:`~HS_Image.set_acquisition_parameters` (optional),
                   :func:`~HS_Image.read_single_frame`

    - ``fprefix``: a file name prefix of the image frame files to be read

      :Type: *string*
      :Required by: (optional) :func:`~HS_Image.set_acquisition_parameters`,
                    (optional) :func:`~HS_Image.read_and_average_frames`

    - ``original_image``: an image read from a file, read and averaged from multiple
      files

      :Type: two-dimensional *ndarray*
      :Required by: :func:`~HS_Image.process_image`,
                    :func:`~HS_Image.locate_bad_pixels`,
                    :func:`~HS_Image.fix_bad_pixels`,
                    :func:`~HS_Image.calculate_sorted_average`,
                    :func:`HS_Centroids.determine_resolution() <HWS.HS_Centroids.HS_Centroids.determine_resolution>`
      :Updated by: :func:`~HS_Image.clear_images`,
                   :func:`~HS_Image.read_single_frame`,
                   :func:`~HS_Image.read_and_average_frames`

    - ``modified_image``: an image obtained by processing ``original_image``

      :Type: *float* or two-dimensional *ndarray* of *floats*
      :Required by: :func:`HS_Centroids.find_centroids_from_image() <HWS.HS_Centroids.HS_Centroids.find_centroids_from_image>`,
                    :func:`HS_Centroids.find_centroids_using_template() <HWS.HS_Centroids.HS_Centroids.find_centroids_using_template>`
      :Updated by: :func:`~HS_Image.process_image`

    - ``background``: The back ground pixel value or an image to be subtracted from
      ``original_image``

      ``background`` is normally obtained from an image acquired without light.

      :Type: non-negative *float* or *ndarray* of non-negative *floats*
      :Required by: :func:`~HS_Image.process_image`

    - ``valid_filesize``: The expected file size of an image file.

      :Type: positive *integer*
      :Required by: :func:`~HS_Image.read_raw`

    - ``bad_pixels``: An array of the x and y coordinates of the pixels tagged to be
      "bad".

      See :func:`~HS_Image.locate_bad_pixels` for an example of how some pixels can
      be tagged to be "bad", and :func:`~HS_Image.fix_bad_pixels` for an example of
      how to deal with them.

      :Type: two-dimensional *ndarray*
      :Requried by: :func:`~HS_Image.fix_bad_pixels`
      :Updated by: :func:`~HS_Image.locate_bad_pixels`

    - ``to_fix_bad_pixels``: a *boolean* to indicate whether to fix bad pixels or not

      If ``True``, :func:`~HS_Image.process_image` will call
      :func:`~HS_Image.fix_bad_pixels` before subtracting the background from
      ``original_image``.

      :Required by: :func:`~HS_Image.process_image`

    - ``threshold_bad_pixels``: a threshold pixel value used to determine which
      pixels are bad

      :func:`~HS_Image.locate_bad_pixels` will tag any pixels whose values are higher
      than ``threshold_bad_pixels`` as bad pixels. The image for this operation must
      be an dark image in most cases.

      If a user specifies the input parameter ``pv_threshold`` when calling
      :func:`~HS_Image.locate_bad_pixels`, the method will use that value and update
      ``threshold_bad_pixels`` as well.

      :Type: non-negative *float*
      :Required by: (optional) :func:`~HS_Image.locate_bad_pixels`
      :Updated by: (optional) :func:`~HS_Image.locate_bad_pixels`

    - ``total_pv``: the sum of all pixel values in ``modified_image``

      :Type: *float*
      :Updated by: :func:`~HS_Image.process_image`

    .. note:: The variables below are currently not in use, and may be removed in
       future revisions unless found useful.

    - ``first_moment``
    - ``second_moment``
    - ``hi_filter_width``
    - ``lo_filter_width``
    - ``hi_filter_power``
    - ``lo_filter_power``
    - ``gpstime``
    - ``max_valid_pv``

    """

    def __init__(
        self,
        location_type=None,
        location=None,
        fprefix=None,
        first_frame=None,
        last_frame=None,
        no_of_frames=None,
        background=0,
        check_status=True,
    ):

        self.location_type = location_type
        self.location = location
        self.first_frame = first_frame
        self.last_frame = last_frame
        self.no_of_frames = no_of_frames
        self.fprefix = fprefix
        self.original_image = None
        self.modified_image = None
        self.background = None
        self.max_valid_pv = None
        self.valid_filesize = 2097152
        self.bad_pixels = None
        self.to_fix_bad_pixels = None
        self.to_mask_image = None
        self.mask_radius = None
        self.mask_center = None
        self.mask_blur = None
        self.threshold_bad_pixels = None
        self.first_moment = None
        self.second_moment = None
        self.hi_filter_width = 10.0
        self.hi_filter_power = 10.0
        self.lo_filter_width = 20.0
        self.lo_filter_power = 10.0
        self.total_pv = None
        self.gpstime = None

        if self.location_type == "file":
            self.no_of_frames = 1

        if (self.location_type == "folder") and (check_status is True):
            self.set_acquisition_parameters()

    def mask_image(self, h=1024, w=1024, center=None, radius=None, blur=None):
        if center is None:  # use the middle of the image
            center = [int(w / 2), int(h / 2)]
        if (
            radius is None
        ):  # use the smallest distance between the center and image walls
            radius = min(center[0], center[1], w - center[0], h - center[1])

        Y, X = ogrid[:h, :w]
        dist_from_center = sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

        mask = dist_from_center <= radius
        mask_out = mask.astype(float)

        mask_out[mask_out > 0] = 1

        from scipy.ndimage.filters import gaussian_filter

        if blur is not None:
            mask_out = gaussian_filter(mask_out, blur)

        return mask_out

    def set_acquisition_parameters(self):
        """Set the various instance variables in preparation for reading the
        image frames.

        This method assigns values to the instance variables based on a user's
        choice when instantiating the class, and raises an exception when there
        is an inconsistency.

        The method checks the following.

        If ``no_of_frames`` is not set by a user:

        - Both ``first_frame`` and ``last_frame`` must be given as positive *integers*.
        - ``first_frame`` must not be larger than ``last_frame``.
        - Then ``no_of_frames`` is set to ``last_frame`` - ``first_frame`` + 1

        If ``no_of_frames`` is set by a user:

        - ``no_of_frames`` must be a positive *integer*.
        - ``first_frame`` can either be ``None`` or a positive *integer*. If ``None``,
          it is set to 1.
        - ``last_frame`` can either be ``None`` or a positive *integer*. If ``None``, it
          is set to ``first_frame`` + ``no_of_frames`` - 1. If not ``None``, the value
          must be equal to that number.

        :Reruires: (optional) ``first_frame``,
                   (optional) ``last_frame``,
                   (optional) ``no_of_frames``
        :Updates: (optional) ``first_frame``,
                  (optional) ``last_frame``,
                  (optional) ``no_of_frames``

        """
        if self.no_of_frames is None:
            if (self.first_frame is None) and (self.last_frame is None):
                raise Exception(
                    "first_frame and last_frame must be set "
                    + "if no_of_frames is not set."
                )

            if self.first_frame is None:
                raise Exception("first_frame must be set if no_of_frames is not set.")

            if self.is_pint(self.first_frame) is False:
                raise Exception("first_frame must be a positive integer.")

            if self.last_frame is None:
                raise Exception("last_frame must be set if no_of_frames is not set.")

            if self.is_pint(self.last_frame) is False:
                raise Exception("last_frame must be a positive integer.")

            if self.first_frame > self.last_frame:
                raise Exception("last_frame must be equal or bigger than first_frame.")
            self.no_of_frames = self.last_frame - self.first_frame + 1

        else:
            if self.is_pint(self.no_of_frames) is False:
                raise Exception("no_of_frames must be a positive integer.")
            if self.first_frame is not None:
                if self.is_pint(self.first_frame) is False:
                    raise Exception("first_frame must be a positive integer.")
            else:
                self.first_frame = 1

            if self.last_frame is not None:
                if self.is_pint(self.first_frame) is False:
                    raise Exception("first_frame must be a positive integer.")

                if self.no_of_frames != self.last_frame - self.first_frame + 1:
                    raise Exception(
                        "first_frame, last_frame, and no_of_frames "
                        + "are not consistent."
                    )
            else:
                self.last_frame = self.first_frame + self.no_of_frames - 1

    def is_pint(self, n):
        """Check if a parameter is a positive *integer*.

        :param n: the parameter to be checked
        :Returns: ``True`` or ``False``
        :rtype: *boolean*

        """
        is_number = isinstance(n, Number)
        if is_number is True:
            return (n == int(n)) and (n > 0)
        else:
            return False

    def read_image(self):
        """Read an image file, or read multiple image files and average them.

        :Requires: ``location``, ``location_type``

        """
        if self.location == None:
            raise Exception("Instance variable location is not set.")

        if self.location_type == "file":
            self.read_single_frame()
        elif self.location_type == "folder":
            self.read_and_average_frames()
        else:
            raise Exception("Instance variable location_type must be file or folder.")

    def clear_images(self):
        """Set ``original_image`` and ``modified_image`` to ``None``.

        :Updates: ``original_image``, ``modified_image``

        """
        self.original_image = None
        self.modified_image = None

    def process_image(self):
        """Process ``original_image`` and set the resulting image to
        ``modified_image``.

        If the instance variable ``to_fix_bad_pixels`` is ``True``, the method calls
        :func:`~HS_Image.fix_bad_pixels` to replace the pixel value of the bad pixels
        with the average of the surrounding pixel values.

        This method then subtracts the background from ``original_image``, and sets any
        pixel values that became negative to zero.

        Finally it sums all the pixel values of ``modified_image`` and sets the value
        to ``total_pv``.

        :Requires: ``background``, ``original_image``
        :Updates: ``modified_image``, ``total_pv``

        """
        if self.original_image is None:
            raise Exception("There is no original_image to process.")

        if self.background is None:
            raise Exception("Instance variable background is not set.")

        tbg = type(self.background)

        if (tbg in [int, float, ndarray]) == False:
            raise Exception(
                "Instance variable background must be a number "
                + "or a numpy array of the same shape as that "
                + "of original_image."
            )

        if tbg == ndarray:
            if (len(self.background) != 1) and (
                self.background.shape != self.original_image.shape
            ):
                raise Exception(
                    "The shape of Instance variable background does not "
                    + "match the shape of original_image."
                )

        if self.to_fix_bad_pixels is True:
            fim = self.fix_bad_pixels()
        else:
            fim = self.original_image

        # ensure that fim is a float ndarray, to further ensure that
        # self.modified_image is also a float ndarray.
        fim = array(fim).astype(float)
        self.modified_image = fim - self.background
        self.modified_image[self.modified_image < 0] = 0

        if self.to_mask_image is True:
            h = self.modified_image.shape[0]
            w = self.modified_image.shape[1]
            mask = self.mask_image(
                h=h,
                w=w,
                center=array(self.mask_center),
                radius=self.mask_radius,
                blur=self.mask_blur,
            )
            self.modified_image = multiply(self.modified_image, mask)

        self.total_pv = sum(self.modified_image)

    def read_single_frame(self):
        """Read a single image frame file whose full path name is gien by the
        instance variable ``location``.

        This method also ensures ``first_frame`` and ``last_frame`` is set to
        ``None``, and ``no_of_frames`` to 1.

        :Requires: ``location``
        :Uses: :func:`~HS_Image.read_raw`
        :Updates: ``first_frame``, ``last_frame``, ``no_of_frames``
                  ``original_image``

        """
        self.first_frame = None
        self.last_frame = None
        self.no_of_frames = 1

        if self.location is None:
            raise Exception(
                "the instance variable location must be the ` + \
                        `full path file name to be read."
            )

        self.original_image = self.read_raw(self.location)

    def read_and_average_frames(self):
        """Read multiple image frame files and average them to make an iamge.

        :Requires: ``location``, ``first_frame``, ``last_frame``, ``no_of_frames``,
                   (optional) ``fprefix``
        :Uses: :func:`~HS_Image.read_raw`
        :Updates: ``no_of_frames``, ``original_image``

        """
        self.set_acquisition_parameters()

        if self.fprefix is None:
            files = [os.path.basename(f) for f in glob.glob(self.location + "*.raw")]
        else:
            files = [
                os.path.basename(f)
                for f in glob.glob(self.location + self.fprefix + "*.raw")
            ]

        if len(files) == 0:
            raise Exception("There are no image files to read.")

        files.sort()
        fpnames = [self.location + f for f in files]

        # Since it is possible to bypass checking mechanism in
        # set_acquisition_parameters, I do provide some flexibility here that
        # the method will still work as long as first_frame and last_frame are
        # given. However, they must still be positive integers.
        if self.is_pint(self.first_frame) is False:
            raise Exception("first_frame must be a positive integer.")

        if self.is_pint(self.last_frame) is False:
            raise Exception("last_frame must be set as a positive integer.")

        self.no_of_frames = abs(self.first_frame - self.last_frame) + 1
        self.original_image = zeros(shape(self.read_raw(fpnames[0])))

        for fpn in fpnames[self.first_frame - 1 : self.last_frame]:
            self.original_image += self.read_raw(fpn)

        self.original_image = self.original_image / float(self.no_of_frames)

    def read_and_average_frames_old(self):
        """An old version of :func:`~HS_Image.read_and_average_frames`

        It is not recommended to use this method, but is here for debugging purposes
        in case something goes wrong.

        May be removed in a future revision.

        """
        if self.fprefix is None:
            files = [os.path.basename(f) for f in glob.glob(self.location + "*.raw")]
            fpnames = [self.location + f for f in files]
        else:
            files = [
                os.path.basename(f)
                for f in glob.glob(self.location + self.fprefix + "*.raw")
            ]

        if len(files) == 0:
            raise Exception("There are no raw files to read.")

        fpnames = [self.location + f for f in files]

        # If first_frame is not defined, assume it is 1
        if self.first_frame is None:
            self.first_frame = 1
        elif type(self.first_frame) != int:
            raise Exception("Instance variable first_frame must be an integer.")

        # If last_frame is not defined, assume it is the last file
        if self.last_frame is None:
            self.last_frame = len(fpnames)
            self.no_of_frames = self.last_frame
        elif type(self.last_frame) != int:
            raise Exception("Instance variable last_frame must be an integer.")

        if self.first_frame > len(fpnames):
            raise Exception(
                "first_frame "
                + str(self.first_frame)
                + " is too big."
                + " There are only "
                + str(len(fpnames))
                + " files avialable."
            )

        if self.last_frame > len(fpnames):
            raise Exception(
                "last_frame "
                + str(self.last_frame)
                + " is too big."
                + " There are only "
                + str(len(fpnames))
                + " files avialable."
            )

        self.no_of_frames = abs(self.first_frame - self.last_frame) + 1
        self.original_image = zeros(shape(self.read_raw(fpnames[1])))

        for fpn in fpnames[self.first_frame - 1 : self.last_frame]:
            self.original_image += self.read_raw(fpn)

        self.original_image = self.original_image / self.no_of_frames

    def read_raw(self, fname):
        """Read a single image frame file.

        This method is device dependent, and is supposed to work with Dalsa
        Pantera 1M60 cameras.

        :param fname: the full path name of the file to be read
        :type fname: *string*
        :Requires: ``valid_filesize``
        :Returns: an image read from the file
        :rtype: two-dimensional *ndarray*

        """
        fsize = int(os.stat(fname).st_size)

        if os.stat(fname).st_size != self.valid_filesize:
            raise Exception(
                "The file size of "
                + fname
                + " is not valid:"
                + "\n The size should be "
                + str(self.valid_filesize)
                + " but was "
                + str(fsize)
                + "."
            )
        # Below is taken from python tutorial:
        # https://docs.python.org/2/library/os.html
        try:
            fp = open(fname)
        except IOError as e:
            print("I/O error({0}): {1}".format(e.errno, e.strerror))
            raise
        else:
            with fp:
                ri = fromfile(fp, uint16).reshape(1024, 1024).astype(float)
                return ri

    def locate_bad_pixels(self, pv_threshold):
        """Locate the pixels whose values exceed ``pv_threshold``

        :param pv_threshold: threshold pixel value
        :type pv_threshold: positive *float*
        :Requires: ``original_image``
        :Updates: ``threshold_bad_pixels``, ``bad_pixels``

        """
        self.bad_pixels = array(where(self.original_image > pv_threshold)).T
        # Convert to xy format from row column by swapping columns
        self.bad_pixels[:, [0, 1]] = self.bad_pixels[:, [1, 0]]
        self.threshold_bad_pixels = pv_threshold

    def fix_bad_pixels(self):
        """Replace the values of the bad pixels with the averaged values of their
        surrounding pixels.

        For each pixel from the instance variable ``bad_pixels``, this method
        calculates the average of the pixel values that surround it, and sets that
        average value as the value of the pixel.

        :Requires: ``original_image``, ``bad_pixels``
        :Returns: an image with bad pixels fixed
        :rtype: two-dimensional *ndarray* of the shape same as that of
                ``original_image``

        """
        fim = self.original_image.copy()
        if self.bad_pixels is None:
            raise Exception(
                "Instance variable bad_pixels is not defined:"
                + "There are no pixels to fix."
            )

        np = shape(self.bad_pixels)[0]
        bp = self.bad_pixels.copy()

        # Convert xy to row column format
        bp[:, [0, 1]] = bp[:, [1, 0]]
        rmax, cmax = shape(fim)
        for k in arange(np):
            lbr = max(0, bp[k, 0] - 1)
            ubr = min(rmax, bp[k, 0] + 1)
            lbc = max(0, bp[k, 1] - 1)
            ubc = min(cmax, bp[k, 1] + 1)
            fpc = 0
            fim[tuple(bp[k])] = 0
            nop = size(fim) - 1
            pv_mean = fim[lbr : ubr + 1, lbc : ubc + 1].sum() / float(nop)
            fim[tuple(bp[k])] = pv_mean

        return fim

    def calculate_sorted_average(self, no_of_pixels=100, image=None):
        """Sort the pixel values, take a number of values from the highest, and
        calculate their average.

        :param no_of_pixels: the number of pixels to take after sorting
        :type no_of_pixels: positive *integer*
        :param image: the image whose pixel values are to be sorted
        :type image: *ndarray*
        :Requires: (optional) ``original_image``
        :Returns: an average of the sorted pixel values
        :rtype: *float*

        """
        if image is None:
            im = self.original_image.ravel()
        else:
            im = image.ravel()

        idx = im.argsort()[::-1][:no_of_pixels]
        return mean(im[idx])
